from autosar.writer.writer_base import ElementWriter
import autosar.base
import autosar.datatype
from fractions import Fraction

class XMLDataTypeWriter(ElementWriter):
    def __init__(self,version, patch):
        super().__init__(version, patch)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {
                              'ArrayDataType': self.writeArrayDataTypeXML,
                              'BooleanDataType': self.writeBooleanDataTypeXML,
                              'IntegerDataType': self.writeIntegerTypeXML,
                              'RealDataType': self.writeRealDataTypeXML,
                              'RecordDataType': self.writeRecordDataTypeXML,
                              'StringDataType': self.writeStringTypeXML,
                              'CompuMethod': self.writeCompuMethodXML,
                              'Unit': self.writeUnitXML,
            }
        elif self.version >= 4.0:
            self.switcher = {
                              'CompuMethod': self.writeCompuMethodXML,
                              'DataConstraint': self.writeDataConstraintXML,
                              'ImplementationDataType': self.writeImplementationDataTypeXML,
                              'SwBaseType': self.writeSwBaseTypeXML,
                              'Unit': self.writeUnitXML,
                              'DataTypeMappingSet': self.writeDataTypeMappingSetXML,
                              'ApplicationPrimitiveDataType': self.writeApplicationPrimitiveDataTypeXML,
                              'ApplicationArrayDataType': self.writeApplicationArrayDataTypeXML,
                              'ApplicationRecordDataType': self.writeApplicationRecordDataTypeXML
            }
        else:
            self.switcher = {}

    def getSupportedXML(self):
        return self.switcher.keys()

    def getSupportedCode(self):
        return []

    def writeElementXML(self, elem):
        xmlWriteFunc = self.switcher.get(type(elem).__name__)
        if xmlWriteFunc is not None:
            return xmlWriteFunc(elem)
        else:
            return None

    def writeElementCode(self, elem, localvars):
        raise NotImplementedError('writeElementCode')

    def writeIntegerTypeXML(self, elem):
        assert isinstance(elem,autosar.datatype.IntegerDataType)
        lines = ["<%s>"%elem.tag(self.version)]
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        if elem.compuMethodRef is not None:
            lines.append(self.indent('<SW-DATA-DEF-PROPS>',1))
            lines.append(self.indent('<COMPU-METHOD-REF DEST="COMPU-METHOD">%s</COMPU-METHOD-REF>'%elem.compuMethodRef,2))
            lines.append(self.indent("</SW-DATA-DEF-PROPS>",1))
        lines.append(self.indent('<LOWER-LIMIT INTERVAL-TYPE="%s">%d</LOWER-LIMIT>'%(elem.minValType, elem.minVal),1))
        lines.append(self.indent('<UPPER-LIMIT INTERVAL-TYPE="%s">%d</UPPER-LIMIT>'%(elem.maxValType, elem.maxVal),1))
        lines.append('</%s>'%elem.tag(self.version))
        return lines

    def writeRecordDataTypeXML(self, elem):
        assert(isinstance(elem,autosar.datatype.RecordDataType))
        ws=elem.rootWS()
        assert(isinstance(ws,autosar.Workspace))
        lines=[]
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        lines.append(self.indent('<ELEMENTS>',1))
        for childElem in elem.elements:
            lines.append(self.indent('<RECORD-ELEMENT>',2))
            lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%childElem.name,3))
            dataType=ws.find(childElem.typeRef, role='DataType')
            if dataType is None:
                raise ValueError('invalid reference: '+childElem.typeRef)
            lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(dataType.tag(self.version),dataType.ref),3))
            lines.append(self.indent('</RECORD-ELEMENT>',2))

        lines.append(self.indent('</ELEMENTS>',1))
        lines.append('</%s>'%elem.tag(self.version))
        return lines

    def writeCompuMethodXML(self, elem):
        assert(isinstance(elem, autosar.datatype.CompuMethod))
        ws=elem.rootWS()
        assert(isinstance(ws,autosar.Workspace))
        lines=[]
        lines.append('<{}>'.format(elem.tag(self.version)))
        lines.append(self.indent('<SHORT-NAME>{}</SHORT-NAME>'.format(elem.name),1))
        if elem.category is not None:
            lines.append(self.indent('<CATEGORY>{}</CATEGORY>'.format(elem.category),1))
        if elem.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(elem.adminData),1))
        if elem.unitRef is not None:
            unit = ws.find(elem.unitRef)
            if unit is None:
                raise autosar.base.InvalidUnitRef(elem.unitRef)
            lines.append(self.indent('<UNIT-REF DEST="{0}">{1}</UNIT-REF>'.format(unit.tag(self.version),elem.unitRef),1))
        if elem.intToPhys is not None:
            tag = 'COMPU-INTERNAL-TO-PHYS'
            lines.extend(self.indent(self._writeComputationXML(ws, elem.intToPhys, tag),1))
        if elem.physToInt is not None:
            tag = 'COMPU-PHYS-TO-INTERNAL'
            lines.extend(self.indent(self._writeComputationXML(ws, elem.physToInt, tag),1))
        lines.append('</{}>'.format(elem.tag(self.version)))
        return lines

    def _writeComputationXML(self, ws, computation, tag):
        assert(isinstance(computation, autosar.datatype.Computation))
        lines=[]
        lines.append('<{}>'.format(tag))
        lines.append(self.indent('<COMPU-SCALES>',1))
        for compuScale in computation.elements:
            lines.extend(self.indent(self._writeCompuScaleXML(ws, compuScale), 2))
        lines.append(self.indent('</COMPU-SCALES>',1))
        if computation.defaultValue is not None:
            lines.append(self.indent('<COMPU-DEFAULT-VALUE>', 1))
            if isinstance(computation.defaultValue, (float, int)):
                lines.append(self.indent('<V>{}</V>'.format(self._numberToString(computation.defaultValue)), 2))
            else:
                lines.append(self.indent('<VT>{}</VT>'.format(str(computation.defaultValue)), 2))
            lines.append(self.indent('</COMPU-DEFAULT-VALUE>', 1))
        lines.append('</{}>'.format(tag))
        return lines

    def _writeCompuScaleXML(self, ws, elem):
        lines = []
        lines.append('<{}>'.format(elem.tag(self.version)))
        if elem.label is not None:
            lines.append(self.indent('<SHORT-LABEL>%s</SHORT-LABEL>'%elem.label, 1))
        if elem.symbol is not None:
            lines.append(self.indent('<SYMBOL>%s</SYMBOL>'%elem.symbol, 1))
        if elem.mask is not None:
            lines.append(self.indent('<MASK>%d</MASK>'%elem.mask, 1))

        if elem.lowerLimit is not None or elem.upperLimit is not None:
            if self.version>=4.0:
                lines.append(self.indent('<LOWER-LIMIT INTERVAL-TYPE="{0}">{1}</LOWER-LIMIT>'.format(elem.lowerLimitType,elem.lowerLimit), 1))
                lines.append(self.indent('<UPPER-LIMIT INTERVAL-TYPE="{0}">{1}</UPPER-LIMIT>'.format(elem.upperLimitType,elem.upperLimit), 1))
            else:
                lines.append(self.indent('<LOWER-LIMIT>%d</LOWER-LIMIT>'%elem.lowerLimit, 1))
                lines.append(self.indent('<UPPER-LIMIT>%d</UPPER-LIMIT>'%elem.upperLimit, 1))
        if elem.offset is not None or elem.numerator is not None or elem.denominator is not None:
            lines.extend(self.indent(self._writeCompuRationalXML(elem),1))
        if elem.textValue is not None:
            lines.extend(self.indent(self._writeCompuConstXML(elem),1))
        lines.append('</{}>'.format(elem.tag(self.version)))
        return lines

    def _writeCompuRationalXML(self, elem):
        lines = []
        lines.append('<COMPU-RATIONAL-COEFFS>')
        lines.append(self.indent('<COMPU-NUMERATOR>', 1))
        lines.append(self.indent('<V>{}</V>'.format(elem.offset), 2))
        lines.append(self.indent('<V>{}</V>'.format(elem.numerator), 2))
        lines.append(self.indent('</COMPU-NUMERATOR>', 1))
        lines.append(self.indent('<COMPU-DENOMINATOR>',1))
        lines.append(self.indent('<V>{}</V>'.format(elem.denominator),2))
        lines.append(self.indent('</COMPU-DENOMINATOR>', 1))
        lines.append('</COMPU-RATIONAL-COEFFS>')
        return lines

    def _writeCompuConstXML(self, elem):
        lines = []
        lines.append('<COMPU-CONST>')
        lines.append(self.indent('<VT>{0}</VT>'.format(elem.textValue), 1))
        lines.append('</COMPU-CONST>')
        return lines

    def writeUnitXML(self, elem):
        lines=[]
        lines.append('<{}>'.format(elem.tag(self.version)))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        if elem.displayName is not None:
            lines.append(self.indent('<DISPLAY-NAME>%s</DISPLAY-NAME>'%elem.displayName,1))
        if self.version >= 4.0:
            if elem.factor is not None:
                lines.append(self.indent('<FACTOR-SI-TO-UNIT>%s</FACTOR-SI-TO-UNIT>'%elem.factor,1))
            if elem.offset is not None:
                lines.append(self.indent('<OFFSET-SI-TO-UNIT>%s</OFFSET-SI-TO-UNIT>'%elem.offset,1))
        lines.append('</{}>'.format(elem.tag(self.version)))
        return lines

    def writeArrayDataTypeXML(self, elem):
        assert(isinstance(elem,autosar.datatype.ArrayDataType))
        ws=elem.rootWS()
        assert(isinstance(ws,autosar.Workspace))

        lines=[]
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        lines.append(self.indent('<ELEMENT>',1))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,2))
        dataType=ws.find(elem.typeRef)
        if dataType is None: raise ValueError('invalid reference: '+elem.typeRef)
        lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(dataType.tag(self.version),dataType.ref),2))
        lines.append(self.indent('<MAX-NUMBER-OF-ELEMENTS>%d</MAX-NUMBER-OF-ELEMENTS>'%elem.length,2))
        lines.append(self.indent('</ELEMENT>',1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeBooleanDataTypeXML(self, elem):
        assert(isinstance(elem,autosar.datatype.BooleanDataType))
        lines=[]
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeRealDataTypeXML(self, elem):
        assert(isinstance(elem,autosar.datatype.RealDataType))
        lines=[]
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        if elem.minValType=="INFINITE":
            lines.append(self.indent('<LOWER-LIMIT INTERVAL-TYPE="INFINITE"></LOWER-LIMIT>',1))
        else:
            if isinstance(elem.minVal, str):
                lines.append(self.indent('<LOWER-LIMIT INTERVAL-TYPE="{0.minValType}">{0.minVal}</LOWER-LIMIT>'.format(elem),1))
            else:
                lines.append(self.indent('<LOWER-LIMIT INTERVAL-TYPE="{0.minValType}">{0.minVal:f}</LOWER-LIMIT>'.format(elem),1))
        if elem.maxValType=="INFINITE":
            lines.append(self.indent('<UPPER-LIMIT INTERVAL-TYPE="INFINITE"></UPPER-LIMIT>',1))
        else:
            if isinstance(elem.maxVal, str):
                lines.append(self.indent('<UPPER-LIMIT INTERVAL-TYPE="{0.maxValType}">{0.maxVal}</UPPER-LIMIT>'.format(elem),1))
            else:
                lines.append(self.indent('<UPPER-LIMIT INTERVAL-TYPE="{0.maxValType}">{0.maxVal:f}</UPPER-LIMIT>'.format(elem),1))
        lines.append(self.indent('<ALLOW-NAN>%s</ALLOW-NAN>'%('true' if elem.hasNaN else 'false'),1))
        lines.append(self.indent('<ENCODING>%s</ENCODING>'%elem.encoding,1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeStringTypeXML(self, elem):
        assert(isinstance(elem,autosar.datatype.StringDataType))
        lines=[]
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        lines.append(self.indent('<ENCODING>%s</ENCODING>'%elem.encoding,1))
        lines.append(self.indent('<MAX-NUMBER-OF-CHARS>%d</MAX-NUMBER-OF-CHARS>'%elem.length,1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeDataConstraintXML(self, elem):
        assert(isinstance(elem,autosar.datatype.DataConstraint))
        ws=elem.rootWS()
        assert(isinstance(ws,autosar.Workspace))

        lines=[]
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        if elem.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(elem.adminData),1))
        lines.append(self.indent('<DATA-CONSTR-RULES>', 1))
        for rule in elem.rules:
            lines.extend(self.indent(self.writeDataConstraintRuleXML(rule, elem.constraintLevel), 2))
        lines.append(self.indent('</DATA-CONSTR-RULES>', 1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeDataConstraintRuleXML(self, rule, constraintLevel):
        lines = []
        lines.append("<DATA-CONSTR-RULE>")
        if isinstance(constraintLevel, int):
            lines.append(self.indent('<{0}>{1}</{0}>'.format('CONSTR-LEVEL', str(constraintLevel)), 1))
        if isinstance(rule, autosar.datatype.InternalConstraint):
            tag_name = 'INTERNAL-CONSTRS'
        elif isinstance(rule, autosar.datatype.PhysicalConstraint):
            tag_name = 'PHYS-CONSTRS'
        else:
            raise NotImplementedError(str(type(rule)))
        lines.append(self.indent('<{}>'.format(tag_name), 1))
        lowerLimit = self._numberToString(rule.lowerLimit)
        upperLimit = self._numberToString(rule.upperLimit)

        lines.append(self.indent('<LOWER-LIMIT INTERVAL-TYPE="{0}">{1}</LOWER-LIMIT>'.format(rule.lowerLimitType, lowerLimit), 2))
        lines.append(self.indent('<UPPER-LIMIT INTERVAL-TYPE="{0}">{1}</UPPER-LIMIT>'.format(rule.lowerLimitType, upperLimit), 2))
        lines.append(self.indent('</{}>'.format(tag_name), 1))
        lines.append("</DATA-CONSTR-RULE>")
        return lines

    def writeImplementationDataTypeXML(self, elem):
        assert(isinstance(elem, autosar.datatype.ImplementationDataType))
        ws=elem.rootWS()
        assert(ws is not None)
        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent("<SHORT-NAME>%s</SHORT-NAME>"%elem.name, 1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        lines.append(self.indent("<CATEGORY>%s</CATEGORY>" % elem.category, 1))
        if elem.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(elem.adminData),1))
        lines.append(self.indent("<SW-DATA-DEF-PROPS>", 1))
        if len(elem.variantProps)>=0:
            lines.extend(self.indent(self.writeSwDataDefPropsVariantsXML(ws, elem.variantProps),2))
        lines.append(self.indent("</SW-DATA-DEF-PROPS>", 1))
        if elem.dynamicArraySizeProfile is not None:
            lines.append(self.indent("<DYNAMIC-ARRAY-SIZE-PROFILE>%s</DYNAMIC-ARRAY-SIZE-PROFILE>"%(str(elem.dynamicArraySizeProfile)), 1))
        if len(elem.subElements)>0:
            lines.append(self.indent("<SUB-ELEMENTS>", 1))
            for subElem in elem.subElements:
                lines.extend(self.indent(self.writeImplementationDataElementXML(ws, subElem),2))
            lines.append(self.indent("</SUB-ELEMENTS>", 1))
        if elem.symbolProps is not None:
            lines.extend(self.indent(self.writeSymbolPropsXML(elem.symbolProps),1))
        if elem.typeEmitter is not None:
            lines.append(self.indent("<TYPE-EMITTER>%s</TYPE-EMITTER>"%(elem.typeEmitter), 1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeSwPointerTargetPropsXML(self, ws, elem):
        assert(isinstance(elem, autosar.base.SwPointerTargetProps))
        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        if elem.targetCategory is not None:
            lines.append(self.indent('<TARGET-CATEGORY>%s</TARGET-CATEGORY>'%(elem.targetCategory),1))
        lines.append(self.indent("<SW-DATA-DEF-PROPS>", 1))
        if len(elem.variants)>=0:
            lines.extend(self.indent(self.writeSwDataDefPropsVariantsXML(ws, elem.variants),2))
        lines.append(self.indent("</SW-DATA-DEF-PROPS>", 1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeSwBaseTypeXML(self, elem):
        assert(isinstance(elem, autosar.datatype.SwBaseType))
        ws = elem.rootWS()
        assert(ws is not None)
        lines = []
        lines.append("<{}>".format(elem.tag(self.version)))
        lines.append(self.indent('<SHORT-NAME>{}</SHORT-NAME>'.format(elem.name),1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        if elem.category is not None:
            lines.append(self.indent('<CATEGORY>{}</CATEGORY>'.format(elem.category),1))
        if elem.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(elem.adminData),1))
        if elem.size is not None:
            lines.append(self.indent('<BASE-TYPE-SIZE>{:d}</BASE-TYPE-SIZE>'.format(elem.size),1))
        if elem.typeEncoding is None and ws.profile.swBaseTypeEncodingDefault is not None:
            lines.append(self.indent('<BASE-TYPE-ENCODING>{}</BASE-TYPE-ENCODING>'.format(str(ws.profile.swBaseTypeEncodingDefault)),1))
        else:
            lines.append(self.indent('<BASE-TYPE-ENCODING>{}</BASE-TYPE-ENCODING>'.format(elem.typeEncoding),1))
        if elem.nativeDeclaration is not None:
            lines.append(self.indent('<NATIVE-DECLARATION>{}</NATIVE-DECLARATION>'.format(elem.nativeDeclaration),1))
        lines.append("</{}>".format(elem.tag(self.version)))
        return lines

    def writeImplementationDataElementXML(self, ws, elem):
        assert(isinstance(elem, autosar.datatype.ImplementationDataTypeElement))
        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None: lines.extend(self.indent(tmp,1))
        if elem.category is not None:
            lines.append(self.indent('<CATEGORY>%s</CATEGORY>'%elem.category,1))
        if elem.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(elem.adminData),1))
        if elem.arraySize is not None:
            lines.append(self.indent('<ARRAY-SIZE>%s</ARRAY-SIZE>'%elem.arraySize,1))
        if elem.arraySizeSemantics is not None:
            lines.append(self.indent('<ARRAY-SIZE-SEMANTICS>%s</ARRAY-SIZE-SEMANTICS>'%elem.arraySizeSemantics,1))
        if len(elem.variantProps)>=0:
            lines.append(self.indent("<SW-DATA-DEF-PROPS>", 1))
            lines.extend(self.indent(self.writeSwDataDefPropsVariantsXML(ws, elem.variantProps),2))
            lines.append(self.indent("</SW-DATA-DEF-PROPS>", 1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeDataTypeMappingSetXML(self, elem):
        assert(isinstance(elem, autosar.datatype.DataTypeMappingSet))
        ws=elem.rootWS()
        assert(ws is not None)

        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        dataTypeMappings = []
        modeRequestMappings = []
        for key in sorted(elem.applicationTypeMap.keys()):
            dataTypeMappings.append(elem.getDataTypeMapping(key))
        for key in sorted(elem.modeRequestMap.keys()):
            modeRequestMappings.append(elem.getModeRequestMapping(key))
        if len(dataTypeMappings) > 0:
            lines.append(self.indent('<DATA-TYPE-MAPS>',1))
            for dataTypeMapping in dataTypeMappings:
                lines.extend(self.indent(self.writeDataTypeMapXML(ws, dataTypeMapping),2))
            lines.append(self.indent('</DATA-TYPE-MAPS>',1))
        if len(modeRequestMappings) > 0:
            lines.append(self.indent('<MODE-REQUEST-TYPE-MAPS>',1))
            for modeRequestMapping in modeRequestMappings:
                lines.extend(self.indent(self.writeModeRequestMapXML(ws, modeRequestMapping),2))
            lines.append(self.indent('</MODE-REQUEST-TYPE-MAPS>',1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeDataTypeMapXML(self, ws, elem):
        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        applicationDataType = ws.find(elem.applicationDataTypeRef)
        if applicationDataType is None:
            raise ValueError('Invalid type reference:' + elem.applicationDataTypeRef)
        implementationDataType = ws.find(elem.implementationDataTypeRef)
        if implementationDataType is None:
            raise ValueError('Invalid type reference:' + elem.implementationDataTypeRef)
        lines.append(self.indent('<APPLICATION-DATA-TYPE-REF DEST="%s">%s</APPLICATION-DATA-TYPE-REF>'%(applicationDataType.tag(self.version), applicationDataType.ref),1))
        lines.append(self.indent('<IMPLEMENTATION-DATA-TYPE-REF DEST="%s">%s</IMPLEMENTATION-DATA-TYPE-REF>'%(implementationDataType.tag(self.version), implementationDataType.ref),1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeModeRequestMapXML(self, ws, elem):
        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        modeDeclarationGroup = ws.find(elem.modeDeclarationGroupRef)
        if modeDeclarationGroup is None:
            raise ValueError('Invalid type reference:' + elem.implementationDataTypeRef)
        implementationDataType = ws.find(elem.implementationDataTypeRef)
        if implementationDataType is None:
            raise ValueError('Invalid type reference:' + elem.implementationDataTypeRef)
        lines.append(self.indent('<IMPLEMENTATION-DATA-TYPE-REF DEST="%s">%s</IMPLEMENTATION-DATA-TYPE-REF>'%(implementationDataType.tag(self.version), implementationDataType.ref),1))
        lines.append(self.indent('<MODE-GROUP-REF DEST="%s">%s</MODE-GROUP-REF>'%(modeDeclarationGroup.tag(self.version), modeDeclarationGroup.ref),1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeApplicationPrimitiveDataTypeXML(self, elem):
        assert(isinstance(elem, autosar.datatype.ApplicationPrimitiveDataType))
        ws=elem.rootWS()
        assert(ws is not None)

        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None:
            lines.extend(self.indent(tmp,1))
        if elem.category is not None:
            lines.append(self.indent('<CATEGORY>%s</CATEGORY>'%elem.category,1))
        if len(elem.variantProps)>=0:
            lines.append(self.indent("<SW-DATA-DEF-PROPS>", 1))
            lines.extend(self.indent(self.writeSwDataDefPropsVariantsXML(ws, elem.variantProps),2))
            lines.append(self.indent("</SW-DATA-DEF-PROPS>", 1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeApplicationArrayDataTypeXML(self, elem):
        assert(isinstance(elem, autosar.datatype.ApplicationArrayDataType))
        ws=elem.rootWS()
        assert(ws is not None)

        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tmp = self.writeDescXML(elem)
        if tmp is not None:
            lines.extend(self.indent(tmp,1))
        if elem.category is not None:
            lines.append(self.indent('<CATEGORY>%s</CATEGORY>'%elem.category,1))
        if len(elem.variantProps)>=0:
            lines.append(self.indent("<SW-DATA-DEF-PROPS>", 1))
            lines.extend(self.indent(self.writeSwDataDefPropsVariantsXML(ws, elem.variantProps),2))
            lines.append(self.indent("</SW-DATA-DEF-PROPS>", 1))
        lines.extend(self.indent(self.writeApplicationArrayDataElementXml(ws, elem.element), 1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeApplicationArrayDataElementXml(self, ws, elem):
        assert(isinstance(elem, autosar.datatype.ApplicationArrayElement))

        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        if elem.name is not None:
            lines.append(self.indent('<SHORT-NAME>{}</SHORT-NAME>'.format(elem.name),1))
        if elem.category is not None:
            lines.append(self.indent('<CATEGORY>{}</CATEGORY>'.format(elem.category),1))
        if elem.typeRef is not None:
            dataType = ws.find(elem.typeRef)
            if dataType is None:
                raise autosar.base.InvalidDataTypeRef(elem.typeRef)
            lines.append(self.indent('<TYPE-TREF DEST="{0}">{1}</TYPE-TREF>'.format(dataType.tag(ws.version), elem.typeRef),1))
        if elem.sizeHandling is not None:
            lines.append(self.indent('<ARRAY-SIZE-HANDLING>{}</ARRAY-SIZE-HANDLING>'.format(elem.sizeHandling),1))
        if elem.sizeSemantics is not None:
            lines.append(self.indent('<ARRAY-SIZE-SEMANTICS>{}</ARRAY-SIZE-SEMANTICS>'.format(elem.sizeSemantics),1))
        if elem.arraySize is not None:
            lines.append(self.indent('<MAX-NUMBER-OF-ELEMENTS>{:d}</MAX-NUMBER-OF-ELEMENTS>'.format(elem.arraySize),1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def writeApplicationRecordDataTypeXML(self, elem):
        assert(isinstance(elem, autosar.datatype.ApplicationRecordDataType))
        ws=elem.rootWS()
        assert(ws is not None)

        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        if elem.name is not None:
            lines.append(self.indent('<SHORT-NAME>{}</SHORT-NAME>'.format(elem.name),1))
        if elem.category is not None:
            lines.append(self.indent('<CATEGORY>{}</CATEGORY>'.format(elem.category),1))
        lines.append(self.indent('<ELEMENTS>', 1))
        for childElem in elem.elements:
            lines.extend(self.indent(self._ApplicationRecordElementXML(ws, childElem), 2))
        lines.append(self.indent('</ELEMENTS>', 1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines

    def _ApplicationRecordElementXML(self, ws, elem):
        assert(isinstance(elem, autosar.datatype.ApplicationRecordElement))

        lines = []
        lines.append("<%s>"%elem.tag(self.version))
        if elem.name is not None:
            lines.append(self.indent('<SHORT-NAME>{}</SHORT-NAME>'.format(elem.name), 1))
        if elem.category is not None:
            lines.append(self.indent('<CATEGORY>{}</CATEGORY>'.format(elem.category),1))
        if elem.typeRef is not None:
            dataType = ws.find(elem.typeRef, role = 'DataType')
            if dataType is None:
                raise autosar.base.InvalidDataTypeRef(elem.typeRef)
            lines.append(self.indent('<TYPE-TREF DEST="{0}">{1}</TYPE-TREF>'.format(dataType.tag(self.version), dataType.ref), 1))
        lines.append("</%s>"%elem.tag(self.version))
        return lines


class CodeDataTypeWriter(ElementWriter):
    def __init__(self, version, patch):
        super().__init__(version, patch)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {
                              'ArrayDataType': self.writeArrayDataTypeCode,
                              'BooleanDataType': self.writeBooleanDataTypeCode,
                             'IntegerDataType': self.writeIntegerTypeCode,
                             'RealDataType': self.writeRealDataTypeCode,
                             'RecordDataType': self.writeRecordDataTypeCode,
                             'StringDataType': self.writeStringTypeCode,
                             'CompuMethodConst': self.writeCompuMethodCode,
                             'CompuMethodRational': self.writeCompuMethodCode,
                             'DataTypeUnitElement': self.writeDataTypeUnitElementCode
            }
        elif self.version >= 4.0:
            self.switcher = {
            }
        else:
            self.switcher = {}

    def getSupportedXML(self):
        return []

    def getSupportedCode(self):
        return self.switcher.keys()

    def writeElementXML(self, elem):
        raise NotImplementedError('writeElementXML')

    def writeElementCode(self, elem, localvars):
        codeWriteFunc = self.switcher.get(type(elem).__name__)
        if codeWriteFunc is not None:
            return codeWriteFunc(elem, localvars)
        else:
            return None

    def writeArrayDataTypeCode(self, dataType, localvars):
        assert(isinstance(dataType,autosar.datatype.ArrayDataType))
        lines=[]
        ws=localvars['ws']
        params=[repr(dataType.name)]
        childType = ws.find(dataType.typeRef)
        if childType is None:
            raise ValueError('invalid reference: '+dataType.typeRef)
        if ws.roles['DataType'] is not None:
            params.append(repr(childType.name))
        else:
            params.append(repr(childType.ref))
        params.append(str(dataType.length))

        if dataType.adminData is not None:
            param = self.writeAdminDataCode(dataType.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)
        if hasattr(dataType, 'desc'):
            lines.append('dataType=package.createArrayDataType(%s)'%(', '.join(params)))
            lines.append('dataType.desc = """%s"""'%dataType.desc)
        else:
            lines.append('package.createArrayDataType(%s)'%(', '.join(params)))
        return lines

    def writeBooleanDataTypeCode(self, dataType, localvars):
        lines=[]
        ws=localvars['ws']
        #name
        params=[repr(dataType.name)]
        if hasattr(dataType, 'desc'):
            lines.append('dataType=package.createBooleanDataType(%s)'%(', '.join(params)))
            lines.append('dataType.desc = """%s"""'%dataType.desc)
        else:
            lines.append('package.createBooleanDataType(%s)'%(', '.join(params)))
        return lines

    def writeIntegerTypeCode(self, dataType, localvars):
        lines=[]
        ws=localvars['ws']
        params=[repr(dataType.name)]

        if dataType.compuMethodRef is not None:
            compuMethod = ws.find(dataType.compuMethodRef)
            if compuMethod is None:
                raise ValueError('invalid reference: '+dataType.compuMethodRef)
            if isinstance(compuMethod, autosar.datatype.CompuMethodConst):
                isUnconventionalType=False
                if (dataType.minVal != 0) or (dataType.maxVal != (len(compuMethod.elements)-1)):
                    isUnconventionalType=True
                else:
                    params2=[]
                    index=0
                    for element in compuMethod.elements:
                        if isinstance(element, autosar.datatype.CompuConstElement):
                            if (element.lowerLimit==index) and (element.upperLimit==index):
                                params2.append(repr(element.textValue))
                            else:
                                isUnconventionalType=True
                                break
                        else:
                            raise ValueError('unsupported value found of type: '+str(type(element)))
                        index+=1
                if isUnconventionalType:
                    params.append(str(dataType.minVal))
                    params.append(str(dataType.maxVal))
                    params2=[]
                    for element in compuMethod.elements:
                        if isinstance(element, autosar.datatype.CompuConstElement):
                            if element.lowerLimit==element.upperLimit:
                                params2.append('(%d, %s)'%(element.lowerLimit, repr(element.textValue)))
                            else:
                                params2.append('(%d, %d, %s)'%(element.lowerLimit, element.upperLimit, repr(element.textValue)))
                        else:
                            raise ValueError('unsupported value found of type: '+str(type(element)))
                text='['+','.join(params2)+']'
                if len(text)>200:
                    #this line is way too long, split it
                    lines.extend(self.writeListCode('valueTableList',params2))
                    params.append('valueTable=valueTableList')
                else:
                    params.append('valueTable='+text)
            elif isinstance(compuMethod, autosar.datatype.CompuMethodRational):
                params.append(str(dataType.minVal))
                params.append(str(dataType.maxVal))
                if len(compuMethod.elements)>1:
                    raise NotImplementedError('CompuMethodRational with multiple elements not implemented')
                elif len(compuMethod.elements)==1:
                    elem =compuMethod.elements[0]
                    #offset
                    params.append('offset='+str(elem.offset))
                    #scaling
                    if elem.denominator=="1":
                        params.append("scaling="+elem.numerator)
                    else:
                        params.append("scaling=%s/%s"%(elem.numerator,elem.numerator))
                    #for now, force float for all scaling factors
                    params.append("forceFloatScaling=True")
                #unit
                if compuMethod.unitRef is not None:
                    unit = ws.find(compuMethod.unitRef, role="Unit")
                    if unit is None:
                        raise ValueError('invalid reference: '+compuMethod.unitRef)
                    if ws.roles['Unit'] is not None:
                        ref = unit.name #use name only
                    else:
                        ref = unit.ref #use full reference
                    params.append('unit='+repr(ref))
            else:
                raise ValueError('unknown type:'+str(type(compuMethod)))
        else:
            params.append(str(dataType.minVal))
            params.append(str(dataType.maxVal))
        if dataType.adminData is not None:
            param = self.writeAdminDataCode(dataType.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)
        if hasattr(dataType, 'desc'):
            lines.append('dataType=package.createIntegerDataType(%s)'%(', '.join(params)))
            lines.append('dataType.desc = """%s"""'%dataType.desc)
        else:
            lines.append('package.createIntegerDataType(%s)'%(', '.join(params)))
        return lines

    def writeRealDataTypeCode(self, dataType, localvars):
        lines=[]
        ws=localvars['ws']
        #name
        params=[repr(dataType.name)]
        #minVal
        if dataType.minVal is None:
            params.append('None')
        else:
            params.append('decimal.Decimal("%s")'%(str(dataType.minVal)))
        #maxVal
        if dataType.maxVal is None:
            params.append('None')
        else:
            params.append('decimal.Decimal("%s")'%(str(dataType.maxVal)))
        #minValType
        if dataType.minValType != 'CLOSED':
            params.append('minValType='+repr(dataType.minValType))
        #maxValType
        if dataType.maxValType != 'CLOSED':
            params.append('maxValType='+repr(dataType.maxValType))
        #hasNan
        if dataType.hasNaN:
            params.append('hasNaN=True')
        #encoding
        if dataType.encoding != 'SINGLE':
            params.append('encoding='+repr(dataType.encoding))
        #adminData
        if dataType.adminData is not None:
            param = self.writeAdminDataCode(dataType.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)
        if hasattr(dataType, 'desc'):
            lines.append('dataType=package.createRealDataType(%s)'%(', '.join(params)))
            lines.append('dataType.desc = """{0}"""'.format(dataType.desc))
        else:
            lines.append('package.createRealDataType(%s)'%(', '.join(params)))
        return lines

    def writeRecordDataTypeCode(self, dataType, localvars):
        lines=[]
        ws=localvars['ws']
        params=[repr(dataType.name)]
        params2=[]
        for elem in dataType.elements:
            childType=ws.find(elem.typeRef)
            if childType is None:
                raise ValueError('invalid reference: '+dataType.typeRef)
            if ws.roles['DataType'] is not None:
                ref = childType.name #use only the name
            else:
                ref = childType.ref #use full reference
            params2.append('(%s,%s)'%(repr(elem.name), repr(ref))) #we use the tuple option since this is most convenient
        if len(dataType.elements)>4:
            lines.extend(self.writeListCode('elementList',params2))
            params.append('elementList')
        else:
            params.append('['+', '.join(params2)+']')
        if hasattr(dataType, 'desc'):
            lines.append('dataType=package.createRecordDataType(%s)'%(', '.join(params)))
            lines.append('dataType.desc = """%s"""'%dataType.desc)
        else:
            lines.append('package.createRecordDataType(%s)'%(', '.join(params)))
        return lines

    def writeStringTypeCode(self, dataType, localvars):
        lines=[]
        ws=localvars['ws']
        #name
        params=[repr(dataType.name)]
        #length
        params.append(str(dataType.length))
        #encoding
        if dataType.encoding != 'ISO-8859-1':
            params.append('encoding='+repr(dataType.encoding))
        if hasattr(dataType, 'desc'):
            lines.append('dataType=package.createStringDataType(%s)'%(', '.join(params)))
            lines.append('dataType.desc = """%s"""'%dataType.desc) #the r will force a raw string to be created
        else:
            lines.append('package.createStringDataType(%s)'%(', '.join(params)))
        return lines

    def writeCompuMethodCode(self, dataType, localvars):
        lines=[]
        ws=localvars['ws']
        return lines

    def CompuMethodRational(self, dataType, localvars):
        lines=[]
        ws=localvars['ws']
        return lines

    def writeDataTypeUnitElementCode(self, dataType, localvars):
        lines=[]
        ws=localvars['ws']
        return lines

    def writeImplementationDataTypeCode(self, dataType, localvars):
        lines = []
        return lines

    def writeDataConstraintCode(self, dataType, localvars):
        lines = []
        return lines



import autosar.util.dcf

def importDcf(filename, external = True):
    """
    Convenience method for importing a DCF file into a newly created workspace
    """
    parser = autosar.util.dcf.DcfParser()
    dcf = parser.parse(filename)
    ws = autosar.workspace()
    dcf.loadReferences(ws, external)
    return ws

def createDcf(ws):
    """
    Convenience method for creating a new Dcf instance based on existing workspace
    """
    return autosar.util.dcf.Dcf(ws)



import autosar.component
import autosar.behavior
import autosar.element
import autosar.portinterface
import autosar.datatype
import autosar.mode
import autosar.builder
import copy
import autosar.base
import re
from fractions import Fraction
import collections
import decimal
import sys

_supress_warnings = True

class Package(object):
    packageName = None
    def __init__(self, name, parent=None, role=None):
        self.name = name
        self.elements = []
        self.subPackages = []
        self.parent=parent
        self.role=role
        self.map={'elements':{}, 'packages':{}}
        self.unhandledParser = set() #[PackageParser] unhandled
        self.unhandledWriter =set() #[PackageWriter] Unhandled

    def __getitem__(self,key):
        if isinstance(key,str):
            return self.find(key)
        else:
            raise ValueError('expected string')

    @property
    def ref(self):
        if self.parent is not None:
            return self.parent.ref+'/%s'%self.name
        else:
            return ''

    def find(self,ref):
        if ref.startswith('/'): return self.parent.find(ref)
        ref = ref.partition('/')
        name = ref[0]
        if name in self.map['packages']:
            package=self.map['packages'][name]
            if len(ref[2])>0:
                return package.find(ref[2])
            else:
                return package
        if name in self.map['elements']:
            elem=self.map['elements'][name]
            if len(ref[2])>0:
                return elem.find(ref[2])
            else:
                return elem
        return None

    def findall(self,ref):
        """
        experimental find-method that has some rudimentary support for globs.
        """
        if ref is None: return None
        if ref[0]=='/': ref=ref[1:] #removes initial '/' if it exists
        ref = ref.partition('/')
        if ref[0]=='*' and len(ref[2])==0:
            result=list(self.elements)
            result.extend(self.subPackages)
        else:
            result=[]
            for item in (self.elements+self.subPackages):
                if item.name == ref[0] or ref[0]=='*':
                    if len(ref[2])>0:
                        result.extend(item.findall(ref[2]))
                    else:
                        result.append(item)
            if (len(result)==0) and ('*' in ref[0]):
                p = re.compile(ref[0].replace('*','.*'))
                for item in (self.elements+self.subPackages):
                    m = p.match(item.name)
                    if m is not None:
                        if len(ref[2])>0:
                            result.extend(item.findall(ref[2]))
                        else:
                            result.append(item)
        return result

    def dir(self,ref=None,_prefix=''):
        if ref==None:
            return [_prefix+x.name for x in self.subPackages]+[_prefix+x.name for x in self.elements]
        else:
            ref = ref.partition('/')
            result=self.find(ref[0])
            if result is not None:
                return result.dir(ref[2] if len(ref[2])>0 else None,_prefix+ref[0]+'/')
            else:
                return None

    def delete(self, ref):
        if ref is None: return
        if ref[0]=='/': ref=ref[1:] #removes initial '/' if it exists
        ref = ref.partition('/')
        for i,element in enumerate(self.elements):
            if element.name == ref[0]:
                if len(ref[2])>0:
                    return element.delete(ref[2])
                else:
                    del self.elements[i]
                    del self.map['elements'][ref[0]]
                    break

    def createSenderReceiverInterface(self, name, dataElements=None, modeGroups=None, isService=False, serviceKind = None, adminData=None):
        """
        creates a new sender-receiver port interface. dataElements can either be a single instance of DataElement or a list of DataElements.
        The same applies to modeGroups. isService must be boolean
        """

        ws = self.rootWS()
        assert(ws is not None)

        portInterface = autosar.portinterface.SenderReceiverInterface(str(name), isService, adminData=adminData)
        if dataElements is not None:
            if isinstance(dataElements,collections.abc.Iterable):
                for elem in dataElements:
                    dataType=ws.find(elem.typeRef, role='DataType')
                    if dataType is None:
                        raise ValueError('invalid type reference: '+elem.typeRef)
                    elem.typeRef=dataType.ref #normalize reference to data element
                    portInterface.append(elem)
            elif isinstance(dataElements,autosar.portinterface.DataElement):
                dataType=ws.find(dataElements.typeRef, role='DataType')
                if dataType is None:
                    raise ValueError('invalid type reference: '+dataElements.typeRef)
                dataElements.typeRef=dataType.ref #normalize reference to data element
                portInterface.append(dataElements)
            else:
                raise ValueError("dataElements: expected autosar.portinterface.DataElement instance or list")
        if modeGroups is not None:
            if isinstance(modeGroups,collections.abc.Iterable):
                for elem in modeGroups:
                    portInterface.append(elem)
            elif isinstance(modeGroups,autosar.mode.ModeGroup):
                portInterface.append(modeGroups)
            else:
                raise ValueError("dataElements: expected autosar.portinterface.DataElement instance or list")
        self.append(portInterface)
        return portInterface

    def createParameterInterface(self, name, parameters=None, modeDeclarationGroups=None, isService=False, adminData=None):
        """
        Creates a new parameter port interface. parameter can either be a single instance of Parameter or a list of Parameters.
        The same applies to modeDeclarationGroups. isService must be boolean.
        In a previous version of this function the class DataElement was used instead of Parameter.
        In order to be backward compatible with old code, this method converts from old to new datatype internally
        """
        ws = self.rootWS()
        assert(ws is not None)

        if isinstance(adminData, dict):
            adminDataObj=ws.createAdminData(adminData)
        else:
            adminDataObj = adminData
        if (adminDataObj is not None) and not isinstance(adminDataObj, autosar.base.AdminData):
            raise ValueError("adminData must be of type dict or AdminData")
        portInterface = autosar.portinterface.ParameterInterface(str(name), adminData=adminDataObj)
        if parameters is not None:
            if isinstance(parameters,collections.abc.Iterable):
                for elem in parameters:
                    dataType=ws.find(elem.typeRef, role='DataType')
                    #normalize reference to data element
                    if dataType is None:
                        raise ValueError('invalid type reference: '+elem.typeRef)
                    elem.typeRef=dataType.ref
                    if isinstance(autosar.portinterface.DataElement):
                        #convert into Parameter
                        parameter = autosar.element.ParameterDataPrototype(elem.name, elem.typeRef, elem.swAddressMethodRef, adminData=elem.adminData)
                    else:
                        parameter = elem
                    portInterface.append(parameter)
            elif isinstance(parameters, autosar.portinterface.DataElement):
                dataType=ws.find(parameters.typeRef, role='DataType')
                #normalize reference to data element
                if dataType is None:
                    raise ValueError('invalid type reference: '+parameters.typeRef)
                parameters.typeRef=dataType.ref
                parameter = autosar.element.ParameterDataPrototype(parameters.name, parameters.typeRef,
                                                            parameters.swAddressMethodRef, adminData=parameters.adminData)
                portInterface.append(parameter)
            elif isinstance(parameters, autosar.element.ParameterDataPrototype):
                dataType=ws.find(parameters.typeRef, role='DataType')
                #normalize reference to data element
                if dataType is None:
                    raise ValueError('invalid type reference: '+parameters.typeRef)
                parameters.typeRef=dataType.ref
                portInterface.append(parameters)
            else:
                raise ValueError("parameters: Expected instance of autosar.element.ParameterDataPrototype or list")
        self.append(portInterface)
        return portInterface

    def createModeSwitchInterface(self, name, modeGroup = None, isService=False, adminData=None):
        portInterface = autosar.portinterface.ModeSwitchInterface(name, isService, self, adminData)
        if modeGroup is not None:
            if isinstance(modeGroup, autosar.mode.ModeGroup):
                ws = self.rootWS()
                assert (ws is not None)
                modeDeclarationGroup = ws.find(modeGroup.typeRef, role='ModeDclrGroup')
                if modeDeclarationGroup is None:
                    raise ValueError('invalid type reference: '+modeGroup.typeRef)
                modeGroup.typeRef=modeDeclarationGroup.ref #normalize reference string
                portInterface.modeGroup = modeGroup
                modeGroup.parent = portInterface
            else:
                raise ValueError('modeGroup must be an instance of autosar.mode.ModeGroup or None')
        self.append(portInterface)
        return portInterface

    def createNvDataInterface(self, name, nvDatas=None, isService=False, serviceKind = None, adminData=None):
        """
        creates a new nv-data port interface. nvDatas can either be a single instance of DataElement or a list of DataElements.
        isService must be boolean
        """

        ws = self.rootWS()
        assert(ws is not None)

        portInterface = autosar.portinterface.NvDataInterface(str(name), isService=isService, serviceKind=serviceKind, adminData=adminData)
        if nvDatas is not None:
            if isinstance(nvDatas,collections.abc.Iterable):
                for elem in nvDatas:
                    dataType=ws.find(elem.typeRef, role='DataType')
                    if dataType is None:
                        raise ValueError('invalid type reference: '+elem.typeRef)
                    elem.typeRef=dataType.ref #normalize reference to data element
                    portInterface.append(elem)
            elif isinstance(nvDatas,autosar.portinterface.DataElement):
                dataType=ws.find(nvDatas.typeRef, role='DataType')
                if dataType is None:
                    raise ValueError('invalid type reference: '+nvDatas.typeRef)
                nvDatas.typeRef=dataType.ref #normalize reference to data element
                portInterface.append(nvDatas)
            else:
                raise ValueError("dataElements: expected autosar.portinterface.DataElement instance or list")
        self.append(portInterface)
        return portInterface

    def createSubPackage(self, name, role=None):
        pkg = Package(name)
        self.append(pkg)
        if role is not None:
            ws = self.rootWS()
            assert(ws is not None)
            ws.setRole(pkg.ref, role)
        return pkg

    def rootWS(self):
        if self.parent is None:
            return None
        else:
            return self.parent.rootWS()

    def append(self,elem):
        """appends elem to the self.elements list"""
        isNewElement = True
        if elem.name in self.map['elements']:
            isNewElement = False
            existingElem = self.map['elements'][elem.name]
            if type(elem) != type(existingElem):
                raise TypeError('Error: element %s %s already exists in package %s with different type from new element %s'%(str(type(existingElem)), existingElem.name, self.name, str(type(elem))))
            else:
                if elem != existingElem:
                    raise ValueError('Error: element %s %s already exist in package %s using different definition'%(existingElem.name, str(type(existingElem)), self.name))
        if isNewElement:
            if isinstance(elem,autosar.element.Element):
                self.elements.append(elem)
                elem.parent=self
                self.map['elements'][elem.name]=elem
            elif isinstance(elem,Package):
                self.subPackages.append(elem)
                elem.parent=self
                self.map['packages'][elem.name]=elem
            else:
                raise ValueError('unexpected value type %s'%str(type(elem)))

    def update(self,other):
        """copies/clones each element from other into self.elements"""
        if type(self) == type(other):
            for otherElem in other.elements:
                newElem=copy.deepcopy(otherElem)
                assert(newElem is not None)
                try:
                    i=self.index('elements',otherElem.name)
                    oldElem=self.elements[i]
                    self.elements[i]=newElem
                    oldElem.parent=None
                except ValueError:
                    self.elements.append(newElem)
                newElem.parent=self
        else:
            raise ValueError('cannot update from object of different type')

    def index(self,container,name):
        if container=='elements':
            lst=self.elements
        elif container=='subPackages':
            lst=self.subPackages
        else:
            raise KeyError("%s not in %s"%(container,self.__class__.__name__))
        return autosar.base.indexByName(lst,name)


    def createApplicationSoftwareComponent(self, swcName, behaviorName=None, implementationName=None, multipleInstance=False, autoCreatePortAPIOptions=True):
        """
        Creates a new ApplicationSoftwareComponent object and adds it to the package.
        It also creates an InternalBehavior object as well as an SwcImplementation object.

        """
        ws = self.rootWS()
        assert(ws is not None)
        swc = autosar.component.ApplicationSoftwareComponent(swcName,self)
        self.append(swc)
        self._createInternalBehavior(ws, swc, behaviorName, multipleInstance, autoCreatePortAPIOptions)
        self._createImplementation(swc, implementationName)
        return swc

    def createServiceComponent(self, swcName, behaviorName=None, implementationName=None, multipleInstance=False, autoCreatePortAPIOptions=True):
        """
        Creates a new ApplicationSoftwareComponent object and adds it to the package.
        It also creates an InternalBehavior object as well as an SwcImplementation object.
        """
        ws = self.rootWS()
        assert(ws is not None)

        swc = autosar.component.ServiceComponent(swcName,self)
        self.append(swc)
        self._createInternalBehavior(ws, swc, behaviorName, multipleInstance, autoCreatePortAPIOptions)
        self._createImplementation(swc, implementationName)
        return swc

    def createComplexDeviceDriverComponent(self,swcName,behaviorName=None,implementationName=None,multipleInstance=False, autoCreatePortAPIOptions=True):
        ws = self.rootWS()
        assert(ws is not None)
        swc=autosar.component.ComplexDeviceDriverComponent(swcName, parent=self)
        self.append(swc)
        self._createInternalBehavior(ws, swc, behaviorName, multipleInstance, autoCreatePortAPIOptions)
        self._createImplementation(swc, implementationName)
        return swc

    def createNvBlockComponent(self,swcName,behaviorName=None,implementationName=None,multipleInstance=False, autoCreatePortAPIOptions=False):
        """
        Creates a new NvBlockComponent object and adds it to the package.
        It also creates an InternalBehavior object as well as an SwcImplementation object.

        """
        ws = self.rootWS()
        assert(ws is not None)
        swc = autosar.component.NvBlockComponent(swcName,self)
        self.append(swc)
        self._createInternalBehavior(ws, swc, behaviorName, multipleInstance, autoCreatePortAPIOptions)
        self._createImplementation(swc, implementationName)
        return swc

    def createCompositionComponent(self, componentName, adminData=None):
        component = autosar.component.CompositionComponent(str(componentName), self)
        self.append(component)
        return component

    def _createInternalBehavior(self, ws, swc, behaviorName, multipleInstance, autoCreatePortAPIOptions):
        """
        Initializes swc.behavior object
        For AUTOSAR3, an instance of InternalBehavior is created
        For AUTOSAR4, an instance of SwcInternalBehavior is created
        """
        if behaviorName is None:
            behaviorName = swc.name+'_InternalBehavior'
        if ws.version < 4.0:
            # In AUTOSAR 3.x the internal behavior is a sub-element of the package.
            internalBehavior = autosar.behavior.InternalBehavior(behaviorName,swc.ref,multipleInstance,self)
        else:
            # In AUTOSAR 4.x the internal behavior is a sub-element of the swc.
            internalBehavior = autosar.behavior.SwcInternalBehavior(behaviorName,swc.ref,multipleInstance, swc)
        internalBehavior.autoCreatePortAPIOptions=autoCreatePortAPIOptions
        swc.behavior=internalBehavior
        if ws.version < 4.0:
            # In AUTOSAR 3.x the internal behavior is a sub-element of the package.
            self.append(internalBehavior)

    def _createImplementation(self, swc, implementationName):

        if implementationName is None:
            implementationName = swc.name+'_Implementation'
        swc.implementation = autosar.component.SwcImplementation(implementationName, swc.behavior.ref, parent=self)

        self.append(swc.implementation)


    def createModeDeclarationGroup(self, name, modeDeclarations = None, initialMode = None, category=None, adminData=None):
        """
        creates an instance of autosar.portinterface.ModeDeclarationGroup
        name: name of the ModeDeclarationGroup
        modeDeclarations: list of strings where each string is a mode name. It can also be a list of tuples of type (int, str).
        initialMode: string with name of the initial mode (must be one of the strings in modeDeclarations list)
        category: optional category string
        adminData: optional adminData object (use ws.createAdminData() as constructor)
        """
        ws = self.rootWS()
        assert(ws is not None)
        if isinstance(adminData, dict):
            adminDataObj=ws.createAdminData(adminData)
        else:
            adminDataObj = adminData
        if (adminDataObj is not None) and not isinstance(adminDataObj, autosar.base.AdminData):
            raise ValueError("adminData must be of type dict or AdminData")
        group = autosar.mode.ModeDeclarationGroup(name, None, None, category, self,adminDataObj)
        if modeDeclarations is not None:
            for declaration in modeDeclarations:
                if isinstance(declaration, str):
                    declarationName = declaration
                    item=autosar.mode.ModeDeclaration(declarationName, parent = group)
                elif isinstance(declaration, tuple):
                    declarationValue = declaration[0]
                    declarationName = declaration[1]
                    assert(isinstance(declarationValue, int))
                    assert(isinstance(declarationName, str))
                    item=autosar.mode.ModeDeclaration(declarationName, declarationValue, group)
                else:
                    raise NotImplementedError(type(declaration))
                group.modeDeclarations.append(item)
                if (initialMode is not None) and (declarationName == initialMode):
                    group.initialModeRef = item.ref
            if (initialMode is not None) and (group.initialModeRef is None):
                raise ValueError('initalMode "%s" not a valid modeDeclaration name'%initialMode)
            self.append(group)
        return group

    def createClientServerInterface(self, name, operations, errors=None, isService=False, serviceKind=None, adminData=None):
        """
        creates a new client server interface in current package
        name: name of the interface (string)
        operations: names of the operations in the interface (list of strings)
        errors: Possible errors that can be returned. Can be a single instance of ApplicationError or a list of ApplicationError
        isService: True if this interface is a service interface (bool)
        adminData: optional admindata (dict or autosar.base.AdminData object)
        """
        portInterface = autosar.portinterface.ClientServerInterface(name, isService, serviceKind, self, adminData)
        for name in operations:
            portInterface.append(autosar.portinterface.Operation(name))
        if errors is not None:
            if isinstance(errors, collections.abc.Iterable):
                for error in errors:
                    portInterface.append(error)
            else:
                assert( isinstance(errors, autosar.portinterface.ApplicationError))
                portInterface.append(errors)
        self.append(portInterface)
        return portInterface

    def createSoftwareAddressMethod(self, name):
        item = autosar.element.SoftwareAddressMethod(name)
        self.append(item)
        return item

    def createArrayDataType(self, name, typeRef, arraySize, elementName=None, adminData=None):
        """
        AUTOSAR3:
           Creates an ArrayDataType and adds it to current package
        """
        ws = self.rootWS()
        assert(ws is not None)
        if ws.version >= 4.0:
            raise RuntimeError("This method is only valid in AUTOSAR3")
        else:
            if typeRef.startswith('/'):
                dataType = ws.find(typeRef)
            else:
                dataType = ws.find(typeRef, role='DataType')
            if dataType is None:
                raise autosar.base.InvalidDataTypeRef(typeRef)
            newType = autosar.datatype.ArrayDataType(name, typeRef, arraySize, adminData)
            self.append(newType)
            return newType

    def createIntegerDataType(self, name, min=None, max=None, valueTable=None, offset=None, scaling=None, unit=None, baseTypeRef=None, adminData=None, swCalibrationAccess='NOT-ACCESSIBLE', typeEmitter=None, forceFloatScaling=False, dataConstraint=''):
        """
        AUTOSAR3:
        Helper method for creating integer datatypes in a package.
        In order to use this function you must have a subpackage present with role='CompuMethod'.

        AUTOSAR4:
        Helper method for creating implementation datatypes or application datatypes.
        """
        ws=self.rootWS()
        assert(ws is not None)
        if ws.version >= 4.0:
            raise RuntimeError("This method is only valid in AUTOSAR3")
        else:
            compuMethodRef = self._createCompuMethodAndUnitV3(ws, name, min, max, valueTable, None, offset, scaling, unit, forceFloatScaling)
            lowerLimit, upperLimit = min, max
            if compuMethodRef is not None:
                compuMethod = ws.find(compuMethodRef)
                if lowerLimit is None:
                    lowerLimit = compuMethod.intToPhys.lowerLimit
                if upperLimit is None:
                    upperLimit = compuMethod.intToPhys.upperLimit
            newType=autosar.datatype.IntegerDataType(name, lowerLimit, upperLimit, compuMethodRef=compuMethodRef, adminData=adminData)
            assert(newType is not None)
            self.append(newType)
            return newType

    def createRealDataType(self, name, minVal, maxVal, minValType='CLOSED', maxValType='CLOSED', hasNaN=False, encoding='SINGLE', baseTypeRef=None, typeEmitter=None, adminData=None):
        """
        AUTOSAR 4: Creates a new ImplementationDataType
        AUTOSAR 3: Creates a new instance of autosar.datatype.RealDataType and appends it to current package
        """
        ws=self.rootWS()
        assert(ws is not None)
        if ws.version >= 4.0:
            if baseTypeRef is None:
                raise ValueError('baseTypeRef argument must be given to this method')
            if (minVal == '-INFINITE' or minVal == 'INFINITE'): minVal = '-INF'
            if (maxVal == 'INFINITE'): maxVal = 'INF'
            if minVal == '-INF' and minValType == 'CLOSED':
                minValType = 'OPEN' #automatic correction
            if maxVal == 'INF' and maxValType == 'CLOSED':
                maxValType = 'OPEN' #automatic correction
            dataConstraint = self.createInternalDataConstraint(name+'_DataConstr', minVal, maxVal, minValType, maxValType)
            newType = autosar.datatype.ImplementationDataType(name, 'VALUE', typeEmitter=typeEmitter)
            props = autosar.base.SwDataDefPropsConditional(baseTypeRef=baseTypeRef,
                                                               swCalibrationAccess='NOT-ACCESSIBLE',
                                                               dataConstraintRef=dataConstraint.ref)
            newType.variantProps = [props]
        else:
            if ( (minVal == '-INFINITE') or (minVal == '-INF') or (minVal == 'INFINITE') or (minVal == 'INF') ):
                #automatic correction
                minVal = None
                minValType = 'INFINITE'
            if ( (maxVal == 'INF') or (maxVal == 'INFINITE') ):
                #automatic correction
                maxVal = None
                maxValType = 'INFINITE'
            newType=autosar.datatype.RealDataType(name, minVal, maxVal, minValType, maxValType, hasNaN, encoding, self, adminData)
        self.append(newType)
        return newType

    def createRecordDataType(self, name, elements, adminData=None):
        """
        AUTOSAR3: Create a new instance of RecordDataType and appends it to current package
        """
        ws = self.rootWS()
        assert(ws is not None)
        if ws.version >= 4.0:
            raise RuntimeError("This method is only for AUTOSAR3")
        else:
            processed = []
            for elem in elements:
                elemUnitRef = None
                if isinstance(elem, autosar.datatype.RecordTypeElement):
                    processed.append(elem)
                elif isinstance(elem, tuple):
                    elemName = elem[0]
                    elemUnitRef = elem[1]
                elif isinstance(elem, collections.abc.Mapping):
                    elemName = elem['name']
                    elemUnitRef = elem['typeRef']
                else:
                    raise ValueError('element must be either Mapping, RecordTypeElement or tuple')
                if elemUnitRef is not None:
                    if elemUnitRef.startswith('/'):
                        dataType = ws.find(elemUnitRef)
                    else:
                        dataType = ws.find(elemUnitRef, role='DataType')
                    if dataType is None:
                        raise autosar.base.InvalidDataTypeRef(elemUnitRef)
                    elem = autosar.datatype.RecordTypeElement(elemName, dataType.ref)
                    processed.append(elem)
            dataType = autosar.datatype.RecordDataType(name, processed, self, adminData)
            self.append(dataType)
            return dataType

    def createStringDataType(self, name, length, encoding='ISO-8859-1', adminData=None):
        """
        create a new instance of autosar.datatype.StringDataType and appends it to current package
        """
        dataType=autosar.datatype.StringDataType(name, length, encoding, self, adminData)
        self.append(dataType)
        return dataType

    def createBooleanDataType(self, name, adminData=None):
        """
        create a new instance of autosar.datatype.BooleanDataType and appends it to current package
        """
        dataType=autosar.datatype.BooleanDataType(name, self, adminData)
        self.append(dataType)
        return dataType

    def createConstant(self, name, typeRef, initValue, label = '', adminData=None):
        """
        create a new instance of autosar.constant.Constant and appends it to the current package
        """
        ws=self.rootWS()
        assert(ws is not None)
        if typeRef is not None:
            dataType = ws.find(typeRef, role='DataType')
            if dataType is None:
                raise autosar.base.InvalidDataTypeRef(str(typeRef))
        else:
            if ws.version < 4.0:
                raise ValueError('typeRef argument cannot be None')
            else:
                dataType = None
        if ws.version < 4.0:
            return self._createConstantV3(ws, name, dataType, initValue, adminData)
        else:
            return self._createConstantV4(ws, name, dataType, initValue, label, adminData)

    def _createConstantV3(self, ws, name, dataType, initValue, adminData=None):
        """Creates an AUTOSAR 3 Constant"""
        if isinstance(dataType, autosar.datatype.IntegerDataType):
            if not isinstance(initValue, int):
                raise ValueError('initValue: expected type int, got '+str(type(initValue)))
            value=autosar.constant.IntegerValue(name, dataType.ref, initValue)
        elif isinstance(dataType, autosar.datatype.RecordDataType):
            if isinstance(initValue, collections.abc.Mapping) or isinstance(initValue, collections.abc.Iterable):
                pass
            else:
                raise ValueError('initValue: expected type Mapping or Iterable, got '+str(type(initValue)))
            value=self._createRecordValueV3(ws, name, dataType, initValue)
        elif isinstance(dataType, autosar.datatype.ArrayDataType):
            if isinstance(initValue, collections.abc.Iterable):
                pass
            else:
                raise ValueError('initValue: expected type Iterable, got '+str(type(initValue)))
            value=self._createArrayValueV3(ws, name, dataType, initValue)
        elif isinstance(dataType, autosar.datatype.BooleanDataType):
            if isinstance(initValue, bool):
                pass
            elif isinstance(initValue, str) or isinstance(initValue, int):
                initValue=bool(initValue)
            else:
                raise ValueError('initValue: expected type bool or str, got '+str(type(initValue)))
            value=autosar.constant.BooleanValue(name, dataType.ref, initValue)
        elif isinstance(dataType, autosar.datatype.StringDataType):
            if isinstance(initValue, str):
                pass
            else:
                raise ValueError('initValue: expected type str, got '+str(type(initValue)))
            value=autosar.constant.StringValue(name, dataType.ref, initValue)
        elif isinstance(dataType, autosar.datatype.RealDataType):
            if isinstance(initValue, float) or isinstance(initValue, decimal.Decimal) or isinstance(initValue, int):
                pass
            else:
                raise ValueError('initValue: expected type int, float or Decimal, got '+str(type(initValue)))
            raise NotImplementedError("Creating constants from RealDataType not implemented")
        else:
            raise ValueError('unrecognized type: '+str(type(dataType)))
        assert(value is not None)
        constant = autosar.constant.Constant(name, value, adminData=adminData)
        self.append(constant)
        return constant

    def _createRecordValueV3(self, ws, name, dataType, initValue, parent=None):
        value = autosar.constant.RecordValue(name, dataType.ref, parent = parent)
        if isinstance(initValue, collections.abc.Mapping):
            for elem in dataType.elements:
                if elem.name in initValue:
                    v = initValue[elem.name]
                    childType = ws.find(elem.typeRef, role='DataType')
                    if childType is None:
                        raise ValueError('invalid reference: '+str(elem.typeRef))
                    if isinstance(childType, autosar.datatype.IntegerDataType):
                        if not isinstance(v, int):
                            raise ValueError('v: expected type int, got '+str(type(v)))
                        value.elements.append(autosar.constant.IntegerValue(elem.name, childType.ref, v, value))
                    elif isinstance(childType, autosar.datatype.RecordDataType):
                        if isinstance(v, collections.abc.Mapping) or isinstance(v, collections.abc.Iterable):
                            pass
                        else:
                            raise ValueError('v: expected type Mapping or Iterable, got '+str(type(v)))
                        value.elements.append(self._createRecordValueV3(ws, elem.name, childType, v, value))
                    elif isinstance(childType, autosar.datatype.ArrayDataType):
                        if isinstance(v, collections.abc.Iterable):
                            pass
                        else:
                            raise ValueError('v: expected type Iterable, got '+str(type(v)))
                        value.elements.append(self._createArrayValueV3(ws, elem.name, childType, v, value))
                    elif isinstance(childType, autosar.datatype.BooleanDataType):
                        if isinstance(v, bool):
                            pass
                        elif isinstance(v, str) or isinstance(v, int):
                            v=bool(v)
                        else:
                            raise ValueError('v: expected type bool or str, got '+str(type(v)))
                        value.elements.append(autosar.constant.BooleanValue(elem.name, childType.ref, v, value))
                    elif isinstance(childType, autosar.datatype.StringDataType):
                        if isinstance(v, str):
                            pass
                        else:
                            raise ValueError('v: expected type str, got '+str(type(v)))
                        value.elements.append(autosar.constant.StringValue(elem.name, childType.ref, v, value))
                    elif isinstance(childType, autosar.datatype.RealDataType):
                        if isinstance(v, float) or isinstance(v, decimal.Decimal) or isinstance(v, int):
                            pass
                        else:
                            raise ValueError('v: expected type int, float or Decimal, got '+str(type(v)))
                        raise NotImplementedError("Creating constants from RealDataType not implemented")
                    else:
                        raise ValueError('unrecognized type: '+str(type(childType)))
                else:
                    raise ValueError('%s: missing initValue field: %s'%(name, elem.name))
        else:
            raise NotImplementedError(type(initValue))
        return value


    def _createArrayValueV3(self, ws, name, dataType, initValue, parent=None):
        value = autosar.constant.ArrayValue(name, dataType.ref, parent=parent)
        childType = ws.find(dataType.typeRef, role='DataType')
        if childType is None:
            raise ValueError('invalid reference: '+str(dataType.typeRef))
        if isinstance(initValue, collections.abc.Iterable):
            for i in range(dataType.length):
                try:
                    v=initValue[i]
                except IndexError:
                    raise ValueError('%s: too few elements in initValue, expected %d items, got %d'%(name, int(dataType.length), len(initValue)))
                elemName='%s_%d'%(childType.name,i)
                if isinstance(childType, autosar.datatype.IntegerDataType):
                    if not isinstance(v, int):
                        raise ValueError('v: expected type int, got '+str(type(v)))
                    value.elements.append(autosar.constant.IntegerValue(elemName, childType.ref, v, value))
                elif isinstance(childType, autosar.datatype.RecordDataType):
                    if isinstance(v, collections.abc.Mapping) or isinstance(v, collections.abc.Iterable):
                        value.elements.append(self._createRecordValueV3(ws, elemName, childType, v, value))
                    else:
                        raise ValueError('v: expected type Mapping or Iterable, got '+str(type(v)))
                elif isinstance(childType, autosar.datatype.ArrayDataType):
                    if isinstance(v, collections.abc.Iterable):
                        pass
                    else:
                        raise ValueError('v: expected type Iterable, got '+str(type(v)))
                    value.elements.append(self._createArrayValueV3(ws, elemName, childType, v, value))
                elif isinstance(childType, autosar.datatype.BooleanDataType):
                    if isinstance(v, bool):
                        pass
                    elif isinstance(v, str) or isinstance(v, int):
                        v=bool(v)
                    else:
                        raise ValueError('v: expected type bool or str, got '+str(type(v)))
                    value.elements.append(autosar.constant.BooleanValue(elemName, childType.ref, v, value))
                elif isinstance(childType, autosar.datatype.StringDataType):
                    if isinstance(v, str):
                        pass
                    else:
                        raise ValueError('v: expected type str, got '+str(type(v)))
                    value.elements.append(autosar.constant.StringValue(elemName, childType.ref, v, value))
                elif isinstance(childType, autosar.datatype.RealDataType):
                    if isinstance(v, float) or isinstance(v, decimal.Decimal) or isinstance(v, int):
                        pass
                    else:
                        raise ValueError('v: expected type int, float or Decimal, got '+str(type(v)))
                    raise NotImplementedError("Creating constants from RealDataType not implemented")
                else:
                    raise ValueError('unrecognized type: '+str(type(childType)))
        else:
            raise NotImplementedError(type(initValue))
        return value

    def _createConstantV4(self, ws, name, dataType, initValue, label, adminData=None):
        """
        If label argument is an empty string, the label of the created value is the same as
        the name of the constant.
        """
        if isinstance(label, str) and len(label) == 0:
            label = name
        builder = autosar.builder.ValueBuilder()
        if dataType is None:
            value = builder.build(label, initValue)
        else:
            value = builder.buildFromDataType(dataType, initValue, label, ws)
        assert(value is not None)
        constant = autosar.constant.Constant(name, value, parent=self, adminData=adminData)
        self.append(constant)
        return constant


    def createTextValueConstant(self, name, value):
        """AUTOSAR 4 text value constant"""
        constant = autosar.constant.Constant(name, None, self)
        constant.value = autosar.constant.TextValue(name, value, constant)
        self.append(constant)
        return constant

    def createNumericalValueConstant(self, name, value):
        """AUTOSAR 4 numerical value constant"""
        constant = autosar.constant.Constant(name, None, self)
        constant.value = autosar.constant.NumericalValue(name, value, constant)
        self.append(constant)
        return constant

    def createApplicationValueConstant(self, name, swValueCont = None, swAxisCont = None, valueCategory = None, valueLabel = None):
        """
        (AUTOSAR4)
        Creates a new Constant containing a application value specification
        """
        ws=self.rootWS()
        assert(ws is not None)
        constant = autosar.constant.Constant(name, None, self)
        innerValue = autosar.constant.ApplicationValue(valueLabel, swValueCont, swAxisCont, valueCategory, parent = self)
        constant.value = innerValue
        self.append(constant)
        return constant

    def createConstantFromValue(self, name, value):
        """
        (AUTOSAR4)
        Wraps an already created value in a constant object
        """
        if not isinstance(value, autosar.constant.ValueAR4):
            raise ValueError("value argument must inherit from class ValueAR4 (got {})".format(type(value)))
        ws=self.rootWS()
        assert(ws is not None)
        constant = autosar.constant.Constant(name, value, self)
        self.append(constant)
        return constant



    def createInternalDataConstraint(self, name, lowerLimit, upperLimit, lowerLimitType="CLOSED", upperLimitType="CLOSED", adminData = None):
        ws=self.rootWS()
        assert(ws is not None)
        return self._checkAndCreateDataConstraint(ws, name, 'internalConstraint', lowerLimit, upperLimit, lowerLimitType, upperLimitType, adminData)

    def createPhysicalDataConstraint(self, name, lowerLimit, upperLimit, lowerLimitType = 'CLOSED', upperLimitType = 'CLOSED', adminData = None):
        ws=self.rootWS()
        assert(ws is not None)
        return self._checkAndCreateDataConstraint(ws, name, 'physicalConstraint', lowerLimit, upperLimit, lowerLimitType, upperLimitType, adminData)


    def createSwBaseType(self, name, size=None, encoding=None, nativeDeclaration=None, category='FIXED_LENGTH', adminData=None):
        """
        Creates a SwBaseType object
        """
        ws=self.rootWS()
        assert(ws is not None)

        if isinstance(adminData, dict):
            adminDataObj=ws.createAdminData(adminData)
        else:
            adminDataObj = adminData
        if (adminDataObj is not None) and not isinstance(adminDataObj, autosar.base.AdminData):
            raise ValueError("adminData must be of type dict or AdminData")
        baseType = autosar.datatype.SwBaseType(name, size, encoding, nativeDeclaration, category, self, adminData)
        self.append(baseType)
        return baseType

    def createBaseType(self, name, size, encoding=None, nativeDeclaration=None, adminData=None):
        """
        AUTOSAR4

        alias for createSwBaseType
        """
        return self.createSwBaseType(name, size, encoding, nativeDeclaration, adminData)


    def createApplicationPrimitiveDataType(self, name, dataConstraint = '', compuMethod = None, unit = None, swCalibrationAccess = None, category = None, adminData = None):
        """
        AUTOSAR4

        Creates a new ApplicationPrimitiveDataType in current package.

        name: <SHORT-NAME> (str)
        dataConstraint: Name or reference to autosar.datatype.DataConstraint object (or None). Default is to automatically create a constraint based on given compuMethod
        compuMethod: Name or reference to autosar.datatype.CompuMethod object (or None).
        swCalibrationAccess: Sets the SW-CALIBRATION-ACCESS property (str['NOT-ACCESSIBLE', 'READ-ONLY', 'READ-WRITE']).
        category: The category of this element (str)
        """
        ws=self.rootWS()
        assert(ws is not None)

        adminDataObj = self._checkAdminData(adminData)
        unitObj = self._checkAndCreateUnit(ws, unit)
        dataConstraintObj, compuMethodObj = None, None
        if compuMethod is not None:
            compuMethodObj = self._findElementByRole(ws, compuMethod, 'CompuMethod')
            if compuMethodObj is None:
                raise autosar.base.InvalidCompuMethodRef(compuMethod)
        if dataConstraint is not None:
            if len(dataConstraint)==0:
                dataConstraintName = self._createDataConstraintName(ws, name)
                if compuMethodObj is not None:
                    dataConstraintObj = self._checkAndCreateDataConstraintFromCompuMethod(ws, dataConstraintName, compuMethodObj)
            else:
                dataConstraintObj = self._findElementByRole(ws, dataConstraint, 'DataConstraint')
                if dataConstraintObj is None:
                    raise autosar.base.InvalidDataConstraintRef(dataConstraint)

        unitRef = None if unitObj is None else unitObj.ref
        compuMethodRef = None if compuMethodObj is None else compuMethodObj.ref
        dataConstraintRef = None if dataConstraintObj is None else dataConstraintObj.ref

        if swCalibrationAccess is not None or dataConstraintRef is not None or compuMethodRef is not None or unitRef is not None:
            variantProps = autosar.base.SwDataDefPropsConditional(
                swCalibrationAccess = swCalibrationAccess,
                dataConstraintRef = dataConstraintRef,
                compuMethodRef = compuMethodRef,
                unitRef = unitRef
                )
        else:
            variantProps = None
        dataType = autosar.datatype.ApplicationPrimitiveDataType(name, variantProps, category, parent = self, adminData = adminDataObj)
        self.append(dataType)
        return dataType

    def createApplicationArrayDataType(self, name, element, swCalibrationAccess = None, category = 'ARRAY', adminData = None):
        """
        AUTOSAR4

        Creates a new createApplicationArrayDataType in current package.

        name: <SHORT-NAME> (str)
        swCalibrationAccess: Sets the SW-CALIBRATION-ACCESS property (str['NOT-ACCESSIBLE', 'READ-ONLY', 'READ-WRITE']).
        element: <ELEMENT> (autosar.datatype.ApplicationArrayElement)
        category: The category of this element (str). Default='ARRAY'
        """
        ws=self.rootWS()
        assert(ws is not None)

        adminDataObj = self._checkAdminData(adminData)
        if swCalibrationAccess is not None:
            variantProps = autosar.base.SwDataDefPropsConditional(
                swCalibrationAccess = swCalibrationAccess
                )
        else:
            variantProps = None
        dataType = autosar.datatype.ApplicationArrayDataType(name, element, variantProps, category, parent = self, adminData = adminDataObj)
        self.append(dataType)
        return dataType

    def createApplicationRecordDataType(self, name, elements = None, swCalibrationAccess = None, category = 'STRUCTURE', adminData = None):
        """
        (AUTOSAR4)
        Creates a new ImplementationDataType containing sub elements
        """
        ws=self.rootWS()
        assert(ws is not None)
        adminDataObj = self._checkAdminData(adminData)
        if swCalibrationAccess is not None and len(swCalibrationAccess)==0:
            swCalibrationAccess = ws.profile.swCalibrationAccessDefault

        if swCalibrationAccess is not None:
            variantProps = autosar.base.SwDataDefPropsConditional(swCalibrationAccess = swCalibrationAccess)
        else:
            variantProps = None

        dataType = autosar.datatype.ApplicationRecordDataType(name, variantProps = variantProps, category = category, adminData = adminDataObj)
        if elements is not None:
            for element in elements:
                if not isinstance(element, tuple):
                    raise ValueError('element must be a tuple')
                (elemName, elemTypeRef) = element
                elemType = ws.find(elemTypeRef, role='DataType')
                if elemType is None:
                    raise autosar.base.InvalidDataTypeRef(elemTypeRef)
                dataType.createElement(elemName, elemType.ref)
        self.append(dataType)
        return dataType

    def createImplementationDataTypeRef(self, name, implementationTypeRef, lowerLimit = None, upperLimit = None, valueTable = None, bitmask = None, offset = None, scaling = None, unit = None, forceFloat = False, dataConstraint = '', swCalibrationAccess = '', typeEmitter = None, lowerLimitType = None, upperLimitType = None, category = 'TYPE_REFERENCE', adminData = None):
        """
        AUTOSAR4

        Creates an implementation data type that is a reference (typedef) of another implementation data type
        name: name of the new data type
        typeRef: reference to implementation data type
        """
        return self._createImplementationDataTypeInternal(
            name,
            None,
            implementationTypeRef,
            lowerLimit,
            upperLimit,
            valueTable,
            bitmask,
            offset,
            scaling,
            unit,
            forceFloat,
            dataConstraint,
            swCalibrationAccess,
            typeEmitter,
            lowerLimitType,
            upperLimitType,
            category,
            adminData
        )

    def createImplementationDataTypePtr(self, name, baseTypeRef, swImplPolicy=None, category = 'DATA_REFERENCE', targetCategory = 'VALUE', adminData = None):
        """
        Creates an implementation type that is a C-type pointer to another type
        """
        ws=self.rootWS()
        assert(ws is not None)

        adminDataObj = self._checkAdminData(adminData)

        targetProps = autosar.base.SwPointerTargetProps(targetCategory, autosar.base.SwDataDefPropsConditional(baseTypeRef = baseTypeRef, swImplPolicy=swImplPolicy))
        variantProps =  autosar.base.SwDataDefPropsConditional(swPointerTargetProps = targetProps)
        implementationDataType = autosar.datatype.ImplementationDataType(name, variantProps, category = category, parent = self, adminData = adminDataObj)
        self.append(implementationDataType)
        return implementationDataType

    def createImplementationArrayDataType(self, name, implementationTypeRef, arraySize, elementName = None, swCalibrationAccess = '', typeEmitter = None, category = 'ARRAY', targetCategory = 'TYPE_REFERENCE', adminData = None):
        """
        (AUTOSAR4)
        Creates a new ImplementationDataType that references another type as an array
        """
        ws=self.rootWS()
        assert(ws is not None)
        if elementName is None:
            elementName = name
        if  swCalibrationAccess is not None and len(swCalibrationAccess)==0:
            swCalibrationAccess = ws.profile.swCalibrationAccessDefault

        if implementationTypeRef.startswith('/'):
            dataType = ws.find(implementationTypeRef)
        else:
            dataType = ws.find(implementationTypeRef, role='DataType')
        if dataType is None:
            raise autosar.base.InvalidDataTypeRef(implementationTypeRef)

        newType = autosar.datatype.ImplementationDataType(name, category = category, adminData = adminData)
        outerProps = autosar.base.SwDataDefPropsConditional(swCalibrationAccess = swCalibrationAccess)
        newType.variantProps = [outerProps]
        innerProps = autosar.base.SwDataDefPropsConditional(implementationTypeRef = implementationTypeRef)
        subElement = autosar.datatype.ImplementationDataTypeElement(elementName, targetCategory, arraySize, variantProps = innerProps)
        newType.subElements.append(subElement)
        self.append(newType)
        return newType

    def createImplementationDataType(self, name, baseTypeRef, lowerLimit = None, upperLimit = None, valueTable = None, bitmask = None, offset = None, scaling = None, unit = None, forceFloat = False, dataConstraint='', swCalibrationAccess = '', typeEmitter = None, lowerLimitType = None, upperLimitType = None, category='VALUE', adminData = None):
        return self._createImplementationDataTypeInternal(
            name,
            baseTypeRef,
            None,
            lowerLimit,
            upperLimit,
            valueTable,
            bitmask,
            offset,
            scaling,
            unit,
            forceFloat,
            dataConstraint,
            swCalibrationAccess,
            typeEmitter,
            lowerLimitType,
            upperLimitType,
            category,
            adminData
        )
    def _createImplementationDataTypeInternal(self, name, baseTypeRef, implementationTypeRef, lowerLimit, upperLimit, valueTable, bitmask, offset, factor, unit, forceFloat, dataConstraint, swCalibrationAccess, typeEmitter, lowerLimitType, upperLimitType, category, adminData):
        """
        Creates an implementation data type that wraps a base type. Defaults to value type category
        """
        ws=self.rootWS()
        assert(ws is not None)

        lowerLimit = self._convertNumber(lowerLimit)
        upperLimit = self._convertNumber(upperLimit)

        if swCalibrationAccess is not None and len(swCalibrationAccess)==0:
            swCalibrationAccess = ws.profile.swCalibrationAccessDefault

        adminDataObj = self._checkAdminData(adminData)
        unitObj = self._checkAndCreateUnit(ws, unit)

        if implementationTypeRef is not None:
            referencedType = ws.find(implementationTypeRef)
            if referencedType is None:
                raise autosar.base.InvalidDataTypeRef(implementationTypeRef)
        else:
            referencedType = None

        compuMethodObj = self._checkAndCreateCompuMethod(ws, self._createCompuMethodName(ws, name), unitObj, lowerLimit, upperLimit, offset, factor, bitmask, valueTable, referencedType, forceFloat)
        if dataConstraint is None:
            dataConstraintObj = None
        else:
            if not isinstance(dataConstraint, str):
                raise ValueError('dataConstraint argument must be None or str')
            if len(dataConstraint)==0:
            #Automatically create a data constraint on empty string
                if compuMethodObj is not None:
                    dataConstraintObj = self._checkAndCreateDataConstraintFromCompuMethod(ws, self._createDataConstraintName(ws, name), compuMethodObj)
                else:
                    if lowerLimit is None and upperLimit is None:
                        dataConstraintObj = None
                    else:
                        if upperLimit is None:
                            raise ValueError('lowerLimit cannot be None')
                        if upperLimit is None:
                            raise ValueError('upperLimit cannot be None')
                        if lowerLimitType is None:
                            lowerLimitType = 'CLOSED'
                        if upperLimitType is None:
                            upperLimitType = 'CLOSED'
                        dataConstraintObj = self._checkAndCreateDataConstraint(ws, self._createDataConstraintName(ws, name), 'internalConstraint', lowerLimit, upperLimit, lowerLimitType, upperLimitType)
            else:
                if dataConstraint.startswith('/'):
                    dataConstraintObj = ws.find(dataConstraint)
                else:
                    dataConstraintObj = ws.find(dataConstraint, role = 'DataConstraint')
                if dataConstraintObj is None:
                    raise autosar.base.InvalidDataConstraintRef(dataConstraint)

        unitRef = None if unitObj is None else unitObj.ref
        compuMethodRef = None if compuMethodObj is None else compuMethodObj.ref
        dataConstraintRef = None if dataConstraintObj is None else dataConstraintObj.ref

        variantProps = autosar.base.SwDataDefPropsConditional(swCalibrationAccess = swCalibrationAccess,
                                                              compuMethodRef = compuMethodRef,
                                                              dataConstraintRef = dataConstraintRef,
                                                              baseTypeRef = baseTypeRef,
                                                              implementationTypeRef = implementationTypeRef,
                                                              unitRef = unitRef)
        implementationDataType = autosar.datatype.ImplementationDataType(name, variantProps, typeEmitter=typeEmitter, category = category, parent = self, adminData = adminDataObj)
        self.append(implementationDataType)
        return implementationDataType

    def createImplementationRecordDataType(self, name, elements, swCalibrationAccess = '', category = 'STRUCTURE', adminData = None):
        """
        (AUTOSAR4)
        Creates a new ImplementationDataType containing sub elements
        """
        ws=self.rootWS()
        assert(ws is not None)

        if swCalibrationAccess is not None and len(swCalibrationAccess)==0:
            swCalibrationAccess = ws.profile.swCalibrationAccessDefault

        if swCalibrationAccess is not None:
            variantProps = autosar.base.SwDataDefPropsConditional(swCalibrationAccess = swCalibrationAccess)
        else:
            variantProps = None
        dataType = autosar.datatype.ImplementationDataType(name, variantProps, category = category, adminData = adminData)
        for element in elements:
            if not isinstance(element, tuple):
                raise ValueError('element must be a tuple')
            (elementName, elemTypeRef) = element
            elemType = ws.find(elemTypeRef, role='DataType')
            if elemType is None:
                raise autosar.base.InvalidDataTypeRef(elemTypeRef)
            if isinstance(elemType, autosar.datatype.ImplementationDataType):
                elementProps = autosar.base.SwDataDefPropsConditional(implementationTypeRef = elemType.ref)
            elif isinstance(elemType, autosar.datatype.SwBaseType):
                elementProps = autosar.base.SwDataDefPropsConditional(baseTypeRef = elemType.ref)
            else:
                raise NotImplementedError(type(elemType))
            implementationDataTypeElement = autosar.datatype.ImplementationDataTypeElement(elementName, 'TYPE_REFERENCE', variantProps = elementProps)
            dataType.subElements.append(implementationDataTypeElement)
        self.append(dataType)
        return dataType


    def createUnit(self, shortName, displayName = None, offset = None, scaling = None,):
        ws = self.rootWS()
        assert(ws is not None)
        if ws.roles['Unit'] is None:
            unitPackage = self
        else:
            unitPackage = ws.find(ws.roles['Unit'])
        unitElem = self._checkAndCreateUnit(ws, shortName, displayName, scaling, offset, unitPackage)
        return unitElem

    def createCompuMethodLinear(self, name, offset, scaling, lowerLimit = None, upperLimit = None, lowerLimitType = 'CLOSED', upperLimitType = 'CLOSED', unit = None, defaultValue = None, label = 'SCALING', forceFloat = True, category = 'LINEAR', adminData = None):
        """
        Alias for createCompuMethodRational
        """
        return self.createCompuMethodRational(name, offset, scaling, lowerLimit, upperLimit, lowerLimitType, upperLimitType, unit, defaultValue, label, forceFloat, category, adminData)

    def createCompuMethodRationalPhys(self, name, offset, scaling, lowerLimit = None, upperLimit = None, lowerLimitType = 'CLOSED', upperLimitType = 'CLOSED', unit = None, defaultValue = None, label = 'SCALING', forceFloat = False, useIntToPhys=False, usePhysToInt = True, category = 'LINEAR', adminData = None):
        """
        Alias for createCompuMethodRational but creates a PHYSICAL-TO-INTERNAL mapping by default
        """
        return self.createCompuMethodRational(name, offset, scaling, lowerLimit, upperLimit, lowerLimitType, upperLimitType, unit, defaultValue, label, forceFloat, useIntToPhys, usePhysToInt, category, adminData)

    def createCompuMethodRational(self, name, offset, scaling, lowerLimit = None, upperLimit = None, lowerLimitType = 'CLOSED', upperLimitType = 'CLOSED', unit = None, defaultValue = None, label = 'SCALING', forceFloat = False, useIntToPhys=True, usePhysToInt = False, category = 'LINEAR', adminData = None):
        """
        Creates a new CompuMethodRational object and appends it to the package with package-role 'CompuMethod'

        name: <SHORT-NAME> (str)
        offset: offset (int or float)
        scaling: scaling factor (int, float or rational number). This together with offset creates the numerator and denominator.
        lowerLimit: <LOWER-LIMIT> (int). Default = None.
        uppwerLimit: <UPPER-LIMIT> (int). Default = None.
        lowerLimitType: "INTERVAL-TYPE" of lowerLimit, str['OPEN', 'CLOSED']. Default='CLOSED'. Only applies when lowerLimit is not None.
        upperLimitType: "INTERVAL-TYPE" of upperLimit, str['OPEN', 'CLOSED']. Default='CLOSED'. Only applies when lowerLimit is not None.
        unit: Name of the unit (str). Default = None
        defaultValue: default value for the computation (None, int, float, str). Default = None.
        label: Label for the internally created <COMPU-SCALE> (str). Default = 'SCALING'.
        forceFloat: Forces numerator to become a floating point number (bool). Default = False.
        intToPhys: Enable internal-to-physical computation (bool). Default = True.
        physToInt: Enable physical-to-internal computation (bool). Default = False.
        category: <CATEGORY> of the CompuMethodRational (str). Default = 'LINEAR'.
        adminData: <ADMIN-DATA> for <COMPU-METHOD> (dict). Default = None.

        """
        ws = self.rootWS()
        assert(ws is not None)

        if ws.roles['CompuMethod'] is None:
            compuMethodPackage = self
        else:
            compuMethodPackage = ws.find(ws.roles['CompuMethod'])

        unitRef = None
        unitObj = self._checkAndCreateUnit(ws, unit)
        if unitObj is not None:
            unitRef = unitObj.ref

        adminDataObj = self._checkAdminData(adminData)

        compuMethod = autosar.datatype.CompuMethod(name, useIntToPhys, usePhysToInt, unitRef, category, compuMethodPackage, adminDataObj)
        (numerator, denominator) = self._calcNumeratorDenominator(scaling, forceFloat)
        if useIntToPhys:
            compuMethod.intToPhys.createRationalScaling(
                offset,
                numerator,
                denominator,
                lowerLimit,
                upperLimit,
                lowerLimitType,
                upperLimitType,
                label = label
                )
            if defaultValue is not None:
                compuMethod.intToPhys.defaultValue = defaultValue
        elif usePhysToInt:
            compuMethod.physToInt.createRationalScaling(
                offset,
                numerator,
                denominator,
                lowerLimit,
                upperLimit,
                lowerLimitType,
                upperLimitType,
                label = label
                )
            if defaultValue is not None:
                if useIntToPhys:
                    compuMethod.intToPhys.defaultValue = defaultValue
                elif usePhysToInt:
                    compuMethod.physToInt.defaultValue = defaultValue

        compuMethodPackage.append(compuMethod)
        return compuMethod

    def createCompuMethodConst(self, name, valueTable, unit = None, defaultValue = None, category = 'TEXTTABLE', adminData = None):
        useIntToPhys, usePhysToInt = True, False

        ws = self.rootWS()
        assert(ws is not None)

        if ws.roles['CompuMethod'] is None:
            compuMethodPackage = self
        else:
            compuMethodPackage = ws.find(ws.roles['CompuMethod'])
        unitRef = None
        unitObj = self._checkAndCreateUnit(ws, unit)
        if unitObj is not None:
            unitRef = unitObj.ref
        if isinstance(adminData, dict):
            adminData = autosar.base.createAdminData(adminData)

        compuMethod = autosar.datatype.CompuMethod(name, useIntToPhys, usePhysToInt, unitRef, category, compuMethodPackage, adminData)
        compuMethod.intToPhys.createValueTable(valueTable)
        if defaultValue is not None:
            compuMethod.intToPhys.defaultValue = defaultValue

        compuMethodPackage.append(compuMethod)
        return compuMethod

    def createDataTypeMappingSet(self, name, adminData = None):
        dataTypeMappingSet = autosar.datatype.DataTypeMappingSet(name, adminData = adminData)
        self.append(dataTypeMappingSet)
        return dataTypeMappingSet

    def _calcNumeratorDenominator(self, scalingFactor, forceFloat = False):

        if forceFloat:
           (numerator, denominator) = (float(scalingFactor), 1)
        else:
            f=Fraction.from_float(scalingFactor)
            if f.denominator > 10000: #use the float version in case its not a rational number
                (numerator, denominator) = (float(scalingFactor), 1)
            else:
                (numerator, denominator) = (f.numerator, f.denominator)
        return (numerator, denominator)

    def _checkAndCreateCompuMethod(self, ws, name, unitObj, lowerLimit, upperLimit, offset, scaling, bitmaskTable, valueTable, referencedType, forceFloatScaling, useCategory = True, autoLabel = True):
        """
        Returns CompuMethod object from the package with role 'CompuMethod'.
        If no CompuMethod exists with that name it will be created and then returned.
        """
        if name is None:
            return None

        category = None
        computation = autosar.datatype.Computation()
        if bitmaskTable is not None:
            category = 'BITFIELD_TEXTTABLE'
            computation.createBitMask(bitmaskTable)
        elif valueTable is not None:
            category = 'TEXTTABLE'
            computation.createValueTable(valueTable, autoLabel)
        elif offset is not None and scaling is not None:
            category = 'LINEAR'
            (numerator, denominator) = self._calcNumeratorDenominator(scaling, forceFloatScaling)
            if (lowerLimit is None) and (referencedType is not None) and (referencedType.dataConstraintRef is not None):
                constraint = ws.find(referencedType.dataConstraintRef)
                if constraint is None:
                    raise autosar.base.InvalidDataConstraintRef
                lowerLimit = constraint.lowerLimit
            if (upperLimit is None) and (referencedType is not None) and (referencedType.dataConstraintRef is not None):
                constraint = ws.find(referencedType.dataConstraintRef)
                if constraint is None:
                    raise autosar.base.InvalidDataConstraintRef
                upperLimit = constraint.upperLimit
            computation.createRationalScaling(offset, numerator, denominator, lowerLimit, upperLimit)
        if category is None:
            return None #Creating a compu method does not seem necessary

        compuMethodPackage = None
        if ws.roles['CompuMethod'] is not None:
            compuMethodPackage=ws.find(ws.roles['CompuMethod'])
        if compuMethodPackage is None:
            raise RuntimeError("No package found with role='CompuMethod'")
        compuMethodObj = compuMethodPackage.find(name)
        if compuMethodObj is not None: #Element already exists with that name?
            return compuMethodObj
        unitRef = None if unitObj is None else unitObj.ref

        useIntToPhys, usePhysToInt = True, False
        if not useCategory:
            category = None
        compuMethodObj = autosar.datatype.CompuMethod(name, useIntToPhys, usePhysToInt, unitRef, category, compuMethodPackage)
        compuMethodObj.intToPhys = computation
        compuMethodPackage.append(compuMethodObj)
        return compuMethodObj

    def _checkAndCreateDataConstraintFromCompuMethod(self, ws, name, compuMethodObj):
        constraintType = 'internalConstraint'
        if compuMethodObj.category == 'BITFIELD_TEXTTABLE':
            lowerLimit = 0
            tmp = compuMethodObj.intToPhys.upperLimit
            upperLimit = 2**int.bit_length(tmp)-1
        else:
            lowerLimit = compuMethodObj.intToPhys.lowerLimit
            upperLimit = compuMethodObj.intToPhys.upperLimit
        if (lowerLimit is not None) and upperLimit is not None:
            return self._checkAndCreateDataConstraint(ws, name, constraintType, lowerLimit, upperLimit)
        else:
            return None

    def _checkAndCreateDataConstraint(self, ws, name, constraintType, lowerLimit, upperLimit, lowerLimitType = 'CLOSED', upperLimitType = 'CLOSED', adminData = None):
        if name is None:
            return None
        dataConstraintPackage = None
        if ws.roles['DataConstraint'] is not None:
            dataConstraintPackage = ws.find(ws.roles['DataConstraint'])
        if dataConstraintPackage is None:
            raise RuntimeError("No package found with role='DataConstraint'")
        dataConstraintObj = dataConstraintPackage.find(name)
        if dataConstraintObj is not None: #Element already exists with that name?
            return dataConstraintObj
        else:
            return self._createDataConstraintInPackage(dataConstraintPackage, name, constraintType, lowerLimit, upperLimit, lowerLimitType, upperLimitType, adminData)

    def _findElementByRole(self, ws, name, role):
        if name.startswith('/'):
            return ws.find(name)
        else:
            return ws.find(name, role = role)

    def _createDataConstraintInPackage(self, dataConstraintPackage, name, constraintType, lowerLimit, upperLimit, lowerLimitType, upperLimitType, adminData):
        """
        Returns DataConstraint object from the package with role 'DataConstraint'.
        If no DataConstraint exists with that name it will be created and then returned.
        """
        rules=[]

        rules.append({'type': constraintType, 'lowerLimit':lowerLimit, 'upperLimit':upperLimit, 'lowerLimitType':lowerLimitType, 'upperLimitType':upperLimitType})
        constraint = autosar.datatype.DataConstraint(name, rules, parent=self)
        dataConstraintPackage.append(constraint)
        return constraint

    def _checkAdminData(self, adminData):
        if isinstance(adminData, dict):
            adminDataObj=autosar.base.createAdminData(adminData)
        else:
            adminDataObj = adminData
        if (adminDataObj is not None) and not isinstance(adminDataObj, autosar.base.AdminData):
            raise ValueError("adminData must be of type dict or AdminData")
        return adminDataObj

    def _createCompuMethodAndUnitV3(self, ws, name, lowerLimit, upperLimit, valueTable, bitmask, offset, scaling, unit, forceFloatScaling):
        """
        AUTOSAR3:

        If both compuMethod and unit: Returns (compuMethodRef, unitRef)
        Else if compuMethod only: Returns (compuMethodRef, None)
        Else: Returns (None, None)
        """
        semanticsPackage = None
        unitPackage = None
        compuMethodElem = None
        unitElem = None
        assert(ws is not None)
        if ws.roles['CompuMethod'] is not None:
            semanticsPackage=ws.find(ws.roles['CompuMethod'])
            if semanticsPackage is None:
                raise RuntimeError("no package found with role='CompuMethod'")
        if ws.roles['Unit'] is not None:
            unitPackage=ws.find(ws.roles['Unit'])
            if unitPackage is None:
                raise RuntimeError("no package found with role='Unit'")

        if (lowerLimit is None) and (upperLimit is None) and (valueTable is None) and (bitmask is None) and (offset is None) and (scaling is None) and (unit is None):
            return None

        if unit is not None:
            unitElem = self._checkAndCreateUnit(ws, unit, unitPackage = unitPackage)
        compuMethodElem = self._checkAndCreateCompuMethod(ws, self._createCompuMethodName(ws, name), unitElem, lowerLimit, upperLimit, offset, scaling, None, valueTable, None, forceFloatScaling, useCategory = False, autoLabel = False)
        if (compuMethodElem is not None) and (unitElem is not None):
            compuMethodElem.unitRef = unitElem.ref
        return None if compuMethodElem is None else compuMethodElem.ref

    def _checkAndCreateUnit(self, ws, shortName, displayName = None, factor = None, offset = None, unitPackage = None):
        if shortName is None:
            return None
        if unitPackage is None:
            unitPackage = ws.find(ws.roles['Unit'])
            if unitPackage is None:
                raise RuntimeError("no package found with role='Unit'")
        assert(isinstance(unitPackage, autosar.package.Package))
        unitElem = unitPackage.find(shortName)
        if unitElem is None:
            unitElem = self._createUnitInPackage(unitPackage, shortName, displayName, factor, offset)
        return unitElem

    def _createUnitInPackage(self, unitPackage, shortName, displayName, factor, offset):
        if displayName is None:
            displayName = shortName
        unitElem = autosar.datatype.Unit(shortName, displayName, factor, offset)
        unitPackage.append(unitElem)
        return unitElem

    def _convertNumber(self, number):
        """
        Attempts to convert argument to int.
        If that fails it tries to convert to float.
        If that fails it returns the argument as a string.
        """
        retval = None
        if number is not None:
            try:
                retval = int(number)
            except ValueError:
                try:
                    retval = float(number)
                except ValueError:
                    retval = str(number)
        return retval

    def _createCompuMethodName(self, ws, name):
        return name + ws.profile.compuMethodSuffix

    def _createDataConstraintName(self, ws, name):
        return name + ws.profile.dataConstraintSuffix



from autosar.base import hasAdminData,parseAdminDataNode,parseTextNode
import autosar.portinterface
import autosar.base
import autosar.element
import autosar.mode
from autosar.parser.parser_base import ElementParser

class PortInterfacePackageParser(ElementParser):
    def __init__(self, version=3.0):
        super().__init__(version)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {'SENDER-RECEIVER-INTERFACE': self.parseSenderReceiverInterface,
                             'CALPRM-INTERFACE': self.parseCalPrmInterface,
                             'CLIENT-SERVER-INTERFACE': self.parseClientServerInterface
            }
        elif self.version >= 4.0:
            self.switcher = {
               'SENDER-RECEIVER-INTERFACE': self.parseSenderReceiverInterface,
               'PARAMETER-INTERFACE': self.parseParameterInterface,
               'CLIENT-SERVER-INTERFACE': self.parseClientServerInterface,
               'MODE-SWITCH-INTERFACE': self.parseModeSwitchInterface,
               'NV-DATA-INTERFACE': self.parseNvDataInterface
            }

    def getSupportedTags(self):
        return self.switcher.keys()

    def parseElement(self, xmlElement, parent = None):
        parseFunc = self.switcher.get(xmlElement.tag)
        if parseFunc is not None:
            return parseFunc(xmlElement,parent)
        else:
            return None



    def parseSenderReceiverInterface(self, xmlRoot, parent=None):
        assert (xmlRoot.tag == 'SENDER-RECEIVER-INTERFACE')
        (isService, serviceKind, dataElements, modeGroups, invalidationPolicys) = (False, None, None, None, None)
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'IS-SERVICE':
                isService = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'SERVICE-KIND' and self.version >= 4.0:
                serviceKind = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'DATA-ELEMENTS':
                dataElements = self._parseDataElements(xmlElem)
            elif xmlElem.tag == 'MODE-GROUPS' and self.version < 4.0:
                modeGroups = self._parseModeGroups(xmlElem)
            elif xmlElem.tag == 'INVALIDATION-POLICYS' and self.version >= 4.0:
                invalidationPolicys = self._parseInvalidationPolicys(xmlElem)
            else:
                self.defaultHandler(xmlElem)
        if self.name is not None:
            portinterface = autosar.portinterface.SenderReceiverInterface(self.name, isService, serviceKind, parent = parent, adminData = self.adminData)
            if dataElements is not None:
                for dataElement in dataElements:
                    dataElement.parent = portinterface
                    portinterface.dataElements.append(dataElement)
            if modeGroups is not None:
                for modeGroup in modeGroups:
                    modeGroup.parent = portinterface
                    portinterface.modeGroups.append(modeGroups)
            if invalidationPolicys is not None:
                for policy in invalidationPolicys:
                    policy.parent = portinterface
                    portinterface.invalidationPolicies.append(policy)
            self.pop(portinterface)
            return portinterface
        else:
            self.pop()
            return None

    def _parseDataElements(self, xmlRoot):
        dataElements = []
        if self.version >= 4.0:
            parseMethod = self.parseVariableDataPrototype
            dataElemTag = 'VARIABLE-DATA-PROTOTYPE'
        else:
            parseMethod = self._parseDataElementPrototype
            dataElemTag = 'DATA-ELEMENT-PROTOTYPE'
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == dataElemTag:
                dataElem = parseMethod(xmlElem)
                if dataElem is not None:
                    dataElements.append(dataElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return dataElements

    def _parseModeGroups(self, xmlRoot):
        assert(xmlRoot.tag == 'MODE-GROUPS')
        modeGroups = []

        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'MODE-DECLARATION-GROUP-PROTOTYPE':
                self.push()
                for xmlChild in xmlElem.findall('./*'):
                    typeRef = None
                    if xmlChild.tag == 'TYPE-TREF':
                        typeRef = self.parseTextNode(xmlChild)
                    else:
                        self.defaultHandler(xmlChild)
                if self.name is not None and typeRef is not None:
                    modeGroups.append(autosar.portinterface.ModeGroup(self.name, typeRef))
                self.pop()
            else:
                raise NotImplementedError(xmlElem.tag)
        return modeGroups

    def _parseInvalidationPolicys(self, xmlRoot):
        assert(xmlRoot.tag == 'INVALIDATION-POLICYS')
        policyList = []
        for xmlChild in xmlRoot.findall('./*'):
            if xmlChild.tag == 'INVALIDATION-POLICY':
                invalidationPolicy = self._parseInvalidationPolicy(xmlChild)
                if invalidationPolicy is not None:
                    policyList.append(invalidationPolicy)
            else:
                raise NotImplementedError(xmlChild.tag)
        return policyList

    def _parseDataElementPrototype(self, xmlRoot):
        assert(xmlRoot.tag == 'DATA-ELEMENT-PROTOTYPE')
        (typeRef, isQueued) = (None, False)
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'IS-QUEUED':
                isQueued = True if self.parseTextNode(xmlElem) == 'true' else False
            else:
                self.defaultHandler(xmlElem)
        if (self.name is not None) and (typeRef is not None):
            elem = autosar.element.DataElement(self.name, typeRef, isQueued)
            self.pop(elem)
            return elem
        else:
            raise RuntimeError('SHORT-NAME and TYPE-TREF must not be None')


    def _parseInvalidationPolicy(self, xmlRoot):
        (dataElementRef, handleInvalid) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'DATA-ELEMENT-REF':
                dataElementRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'HANDLE-INVALID':
                handleInvalid = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (dataElementRef is not None) and (handleInvalid is not None):
            return autosar.portinterface.InvalidationPolicy(dataElementRef, handleInvalid)
        else:
            raise RuntimeError('DATA-ELEMENT-REF and HANDLE-INVALID must not be None')

    def parseCalPrmInterface(self,xmlRoot,parent=None):
        assert(xmlRoot.tag=='CALPRM-INTERFACE')
        xmlName = xmlRoot.find("./SHORT-NAME")
        if xmlName is not None:
            portInterface = autosar.portinterface.ParameterInterface(xmlName.text)
            if xmlRoot.find("./IS-SERVICE").text == 'true': portInterface.isService = True
            for xmlElem in xmlRoot.findall('./CALPRM-ELEMENTS/CALPRM-ELEMENT-PROTOTYPE'):
                xmlElemName = xmlElem.find("./SHORT-NAME")
                if xmlElemName is not None:
                    typeRef=xmlElem.find("./TYPE-TREF").text
                    parameter = autosar.element.ParameterDataPrototype(xmlElemName.text,typeRef,parent=portInterface)
                    if hasAdminData(xmlElem):
                        parameter.adminData=parseAdminDataNode(xmlElem.find('ADMIN-DATA'))
                    if xmlElem.find('SW-DATA-DEF-PROPS'):
                        for xmlItem in xmlElem.findall('SW-DATA-DEF-PROPS/SW-ADDR-METHOD-REF'):
                            parameter.swAddressMethodRef = self.parseTextNode(xmlItem)
                    portInterface.append(parameter)
            return portInterface

    def parseClientServerInterface(self,xmlRoot,parent=None):
        assert(xmlRoot.tag=='CLIENT-SERVER-INTERFACE')
        name = self.parseTextNode(xmlRoot.find('SHORT-NAME'))
        if name is not None:
            portInterface = autosar.portinterface.ClientServerInterface(name)
            if hasAdminData(xmlRoot):
                portInterface.adminData=parseAdminDataNode(xmlRoot.find('ADMIN-DATA'))
            for xmlElem in xmlRoot.findall('./*'):
                if xmlElem.tag == 'DESC':
                    pass #implement later
                elif (xmlElem.tag == 'SHORT-NAME') or (xmlElem.tag == 'ADMIN-DATA'):
                    continue
                elif xmlElem.tag == 'IS-SERVICE':
                    if self.parseTextNode(xmlElem) == 'true':
                        portInterface.isService = True
                elif xmlElem.tag == 'OPERATIONS':
                    for xmlChildItem in xmlElem.findall('./*'):
                        if (self.version < 4.0 and xmlChildItem.tag == 'OPERATION-PROTOTYPE') or (self.version >= 4.0 and xmlChildItem.tag == 'CLIENT-SERVER-OPERATION'):
                            operation = self._parseOperationPrototype(xmlChildItem, portInterface)
                            portInterface.operations.append(operation)
                        else:
                            raise NotImplementedError(xmlChildItem.tag)
                elif xmlElem.tag == 'POSSIBLE-ERRORS':
                    for xmlError in xmlElem.findall('APPLICATION-ERROR'):
                        applicationError = self._parseApplicationError(xmlError, portInterface)
                        portInterface.applicationErrors.append(applicationError)
                elif xmlElem.tag == 'SERVICE-KIND':
                    portInterface.serviceKind = self.parseTextNode(xmlElem)
                else:
                    raise NotImplementedError(xmlElem.tag)
            return portInterface

    def parseParameterInterface(self,xmlRoot,parent=None):
        (name, adminData, isService, xmlParameters) = (None, None, False, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ADMIN-DATA':
                adminData = self.parseAdminDataNode(xmlElem)
            elif xmlElem.tag == 'IS-SERVICE':
                if self.parseTextNode(xmlElem) == 'true':
                    isService = True
            elif xmlElem.tag == 'PARAMETERS':
                xmlParameters = xmlElem
            else:
                raise NotImplementedError(xmlElem.tag)

        if (name is not None) and (xmlParameters is not None):
            portInterface = autosar.portinterface.ParameterInterface(name, isService, parent, adminData)
            for xmlChild in xmlParameters.findall('./*'):
                if xmlChild.tag == 'PARAMETER-DATA-PROTOTYPE':
                    parameter = self._parseParameterDataPrototype(xmlChild, portInterface)
                    portInterface.append(parameter)
                else:
                    raise NotImplementedError(xmlChild.tag)
            return portInterface


    def parseModeSwitchInterface(self,xmlRoot,parent=None):
        (name, adminData, desc, isService, xmlModeGroup) = (None, None, None, False, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ADMIN-DATA':
                adminData = self.parseAdminDataNode(xmlElem)
            elif xmlElem.tag == 'IS-SERVICE':
                if self.parseTextNode(xmlElem) == 'true':
                    isService = True
            elif xmlElem.tag == 'MODE-GROUP':
                xmlModeGroup = xmlElem
            else:
                raise NotImplementedError(xmlElem.tag)

        if (name is not None) and (xmlModeGroup is not None):
            portInterface = autosar.portinterface.ModeSwitchInterface(name, isService, parent, adminData)
            portInterface.modeGroup=self._parseModeGroup(xmlModeGroup, portInterface)
            return portInterface

    def _parseModeGroup(self, xmlModeGroup, parent):
        if self.version>=4.0:
            assert(xmlModeGroup.tag == "MODE-GROUP")
        else:
            assert(xmlModeGroup.tag == "MODE-DECLARATION-GROUP-PROTOTYPE")
        name = self.parseTextNode(xmlModeGroup.find('SHORT-NAME'))
        typeRef = self.parseTextNode(xmlModeGroup.find('TYPE-TREF'))
        return autosar.mode.ModeGroup(name, typeRef, parent)

    def _parseOperationPrototype(self, xmlOperation, parent):
        (name, xmlDesc, xmlArguments, xmlPossibleErrorRefs) = (None, None, None, None)
        for xmlElem in xmlOperation.findall('./*'):
            if xmlElem.tag == 'ADMIN-DATA':
                pass #implement later
            elif xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'DESC':
                xmlDesc = xmlElem
            elif xmlElem.tag == 'ARGUMENTS':
                xmlArguments = xmlElem
            elif xmlElem.tag == 'POSSIBLE-ERROR-REFS':
                xmlPossibleErrorRefs = xmlElem
            else:
                raise NotImplementedError(xmlElem.tag)

        if name is not None:
            operation = autosar.portinterface.Operation(name, parent)
            argumentTag = 'ARGUMENT-DATA-PROTOTYPE' if self.version >= 4.0 else 'ARGUMENT-PROTOTYPE'
            if xmlDesc is not None:
                self.parseDesc(xmlOperation,operation)
            if xmlArguments is not None:
                for xmlChild in xmlArguments.findall('./*'):
                    if xmlChild.tag == argumentTag:
                        if self.version >= 4.0:
                            argument = self._parseOperationArgumentV4(xmlChild, operation)
                        else:
                            argument = self._parseOperationArgumentV3(xmlChild, operation)
                        operation.arguments.append(argument)
                        argument.parent=operation
                    else:
                        raise NotImplementedError(xmlChild.tag)
            if xmlPossibleErrorRefs is not None:
                for xmlChild in xmlPossibleErrorRefs.findall('./*'):
                    if xmlChild.tag == 'POSSIBLE-ERROR-REF':
                        operation.errorRefs.append(self.parseTextNode(xmlChild))
                    else:
                        raise NotImplementedError(xmlChild.tag)
            return operation

    def _parseOperationArgumentV3(self, xmlArgument, parent):
        (name, typeRef, direction) = (None, None, None)
        for xmlElem in xmlArgument.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'DIRECTION':
                direction = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (name is not None) and (typeRef is not None) and (direction is not None):
            return autosar.portinterface.Argument(name, typeRef, direction)
        else:
            raise RuntimeError('SHORT-NAME, TYPE-TREF and DIRECTION must have valid values')

    def _parseOperationArgumentV4(self, xmlArgument, parent):
        (argument, typeRef, direction, props_variants, serverArgumentImplPolicy) = (None, None, None, None, None)
        self.push()
        for xmlElem in xmlArgument.findall('./*'):
            if xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'DIRECTION':
                direction = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SW-DATA-DEF-PROPS':
                props_variants = self.parseSwDataDefProps(xmlElem)
            elif xmlElem.tag == 'SERVER-ARGUMENT-IMPL-POLICY':
                serverArgumentImplPolicy=self.parseTextNode(xmlElem)
            else:
                self.baseHandler(xmlElem)
        if (self.name is not None) and (typeRef is not None) and (direction is not None):
            argument = autosar.portinterface.Argument(self.name, typeRef, direction, serverArgumentImplPolicy = serverArgumentImplPolicy)
            if props_variants is not None:
                argument.swCalibrationAccess = props_variants[0].swCalibrationAccess
        else:
            raise RuntimeError('SHORT-NAME, TYPE-TREF and DIRECTION must have valid values')
        self.pop(argument)
        return argument

    def _parseApplicationError(self, xmlElem, parent):
        name=self.parseTextNode(xmlElem.find("./SHORT-NAME"))
        errorCode=self.parseTextNode(xmlElem.find("./ERROR-CODE"))
        return autosar.portinterface.ApplicationError(name, errorCode, parent)

    def _parseParameterDataPrototype(self, xmlElem, parent):
        (name, adminData, typeRef, props_variants) = (None, None, None, None)
        for xmlElem in xmlElem.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ADMIN-DATA':
                adminData = self.parseAdminDataNode(xmlElem)
            elif xmlElem.tag == 'SW-DATA-DEF-PROPS':
                props_variants = self.parseSwDataDefProps(xmlElem)
            elif xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)

        if (name is not None) and (typeRef is not None):
            parameter = autosar.element.ParameterDataPrototype(name, typeRef, parent=parent, adminData=adminData)
            if props_variants is not None:
                parameter.swCalibrationAccess = props_variants[0].swCalibrationAccess
                parameter.swAddressMethodRef = props_variants[0].swAddressMethodRef
            return parameter

    def parseNvDataInterface(self, xmlRoot, parent=None):
        assert (xmlRoot.tag == 'NV-DATA-INTERFACE')
        (isService, serviceKind, nvDatas) = (False, None, None)
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'IS-SERVICE':
                isService = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'SERVICE-KIND' and self.version >= 4.0:
                serviceKind = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'NV-DATAS':
                nvDatas = self._parseDataElements(xmlElem)
            else:
                self.defaultHandler(xmlElem)
        if self.name is not None:
            nvDataInterface = autosar.portinterface.NvDataInterface(self.name, isService, serviceKind, parent = parent, adminData = self.adminData)
            if nvDatas is not None:
                for nvData in nvDatas:
                    nvData.parent = nvDataInterface
                    nvDataInterface.nvDatas.append(nvData)
            self.pop(nvDataInterface)
            return nvDataInterface
        else:
            self.pop()
            return None


class SoftwareAddressMethodParser(ElementParser):
    def __init__(self,version=3.0):
        self.version=version

    def getSupportedTags(self):
        return ['SW-ADDR-METHOD']

    def parseElement(self, xmlElement, parent = None):
        if xmlElement.tag == 'SW-ADDR-METHOD':
            return self.parseSWAddrMethod(xmlElement, parent)
        else:
            return None

    def parseSWAddrMethod(self,xmlRoot,rootProject=None,parent=None):
        assert(xmlRoot.tag == 'SW-ADDR-METHOD')
        name = xmlRoot.find("./SHORT-NAME").text
        return autosar.element.SoftwareAddressMethod(name)



import xml.etree.ElementTree as ElementTree
import re

pVersion = re.compile(r"(\d+)\.(\d+)\.(\d+)")

class AdminData:
    def __init__(self):
        self.specialDataGroups = []
    def asdict(self):
        retval={'type': self.__class__.__name__, 'specialDataGroups':[]}
        for elem in self.specialDataGroups:
            retval['specialDataGroups'].append(elem.asdict())
        return retval

    def __eq__(self, other):
        if isinstance(other, self.__class__) and len(self.specialDataGroups) == len(other.specialDataGroups):
            for i,elem in enumerate(self.specialDataGroups):
                if elem != other.specialDataGroups[i]:
                    return False
            return True
        return False

    def __ne__(self, other): return not (self == other)

class SpecialDataGroup(object):
    def __init__(self,SDG_GID,SD=None,SD_GID=None):
        self.SDG_GID=SDG_GID
        self.SD = []
        if SD is not None or SD_GID is not None:
            self.SD.append(SpecialData(SD, SD_GID))

    # def asdict(self):
    #    data = {'type': self.__class__.__name__}
    #    if self.SDG_GID is not None: data['SDG_GID']=self.SDG_GID
    #    if self.SD is not None: data['SD']=self.SD
    #    if self.SD_GID is not None: data['SD_GID']=self.SD_GID
    #    return data

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.SDG_GID == other.SDG_GID:
                for i,SD in enumerate(self.SD):
                    other_SD = other.SD[i]
                    if SD.TEXT != other_SD.TEXT or SD.GID != other_SD.GID:
                        return False
                return True
        return False

    def __ne__(self, other): return not (self == other)

class SpecialData:
    def __init__(self, TEXT, GID):
        self.TEXT = TEXT
        self.GID = GID

def removeNamespace(doc, namespace):
    """Removes XML namespace in place."""
    ns = u'{%s}' % namespace
    nsl = len(ns)
    for elem in doc.iter():
        if elem.tag.startswith(ns):
            elem.tag = elem.tag[nsl:]

def parseXMLFile(filename,namespace=None):
    arxml_tree = ElementTree.ElementTree()
    arxml_tree.parse(filename)
    arxml_root = arxml_tree.getroot()
    if namespace is not None:
        removeNamespace(arxml_root,namespace)
    return arxml_root

def getXMLNamespace(element):
    m = re.match(r'\{(.*)\}', element.tag)
    return m.group(1) if m else None

def splitRef(ref):
    """splits an autosar url string into an array"""
    if isinstance(ref,str):
        if ref[0]=='/': return ref[1:].split('/')
        else: return ref.split('/')
    return None

def hasAdminData(xmlRoot):
    return True if xmlRoot.find('ADMIN-DATA') is not None else False

def parseAdminDataNode(xmlRoot):
    if xmlRoot is None: return None
    assert(xmlRoot.tag=='ADMIN-DATA')
    adminData=AdminData()
    xmlSDGS = xmlRoot.find('./SDGS')
    if xmlSDGS is not None:
        for xmlElem in xmlSDGS.findall('./SDG'):
            GID=xmlElem.attrib['GID']
            SD=None
            SD_GID=None
            xmlSD = xmlElem.find('SD')
            if xmlSD is not None:
                SD=xmlSD.text
                try:
                    SD_GID=xmlSD.attrib['GID']
                except KeyError: pass
            adminData.specialDataGroups.append(SpecialDataGroup(GID,SD,SD_GID))
    return adminData

def parseTextNode(xmlElem):
    return None if xmlElem is None else xmlElem.text
def parseIntNode(xmlElem):
    return None if xmlElem is None else int(xmlElem.text)
def parseFloatNode(xmlElem):
    return None if xmlElem is None else float(xmlElem.text)
def parseBooleanNode(xmlElem):
    return None if xmlElem is None else parseBoolean(xmlElem.text)

def parseBoolean(value):
    if value is None:
        return None
    if isinstance(value,str):
        if value == 'true': return True
        elif value =='false': return False
    raise ValueError(value)

def indexByName(lst,name):
    assert(isinstance(lst,list))
    assert(isinstance(name,str))
    for i,item in enumerate(lst):
        if item.name == name: return i
    raise ValueError('%s not in list'%name)

def createAdminData(data):
    adminData = AdminData()
    SDG_GID = data.get('SDG_GID',None)
    if 'SDG' in data:
        group = SpecialDataGroup(SDG_GID)
        for item in data['SDG']:
            SD_GID = item.get('SD_GID',None)
            SD = item.get('SD',None)
            group.SD.append(SpecialData(SD, SD_GID))
        adminData.specialDataGroups.append(group)
    else:
        SD_GID = data.get('SD_GID',None)
        SD = data.get('SD',None)
        adminData.specialDataGroups.append(SpecialDataGroup(SDG_GID,SD,SD_GID))
    return adminData

def parseAutosarVersionAndSchema(xmlRoot):
    """
    Parses AUTOSAR version from the schemaLocation attribute in the root AUTOSAR tag

    For AUTOSAR versions 4.3 and below (e.g. "http://autosar.org/schema/r4.0 AUTOSAR_4-3-0.xsd")
    Returns a tuple with major, minor, patch, None, schemaFile. Types are (int, int, int, NoneType, str)

    For AUTOSAR versions 4.4 and above (e.g. "http://autosar.org/schema/r4.0 AUTOSAR_00044.xsd")
    Returns a tuple with major, minor, None, release, schemaFile
    This will now report (major,minor) as (4,0) since it will now extract from the "r4.0"-part of the attribute.
    """
    schemaLocation = None
    for key in xmlRoot.attrib.keys():
        if key.endswith('schemaLocation'):
            value = xmlRoot.attrib[key]
            #Retreive the schema file
            result = re.search(r'(^[ ]+\.xsd)', value)
            tmp = value.partition(' ')
            if len(tmp[2])>0:
                schemaFile = tmp[2]
            else:
                schemaFile = None
            #Is this AUTOSAR 3?
            result = re.search(r'(\d)\.(\d)\.(\d)', value)
            if result is not None:
                return (int(result.group(1)), int(result.group(2)), int(result.group(3)), None,  schemaFile)
            else:
                #Is this AUTOSAR 4.0 to 4.3?
                result = re.search(r'(\d)-(\d)-(\d).*\.xsd', value)
                if result is not None:
                    return (int(result.group(1)),int(result.group(2)),int(result.group(3)), None, schemaFile)
                else:
                    #Is this AUTOSAR 4.4 or above?
                    result = re.search(r'r(\d+)\.(\d+)\s+(autosar|AUTOSAR)_(\d+).xsd', value)
                    if result is not None:
                        return (int(result.group(1)),int(result.group(2)),None, int(result.group(4)), schemaFile)

    return (None, None, None, None, None)

def applyFilter(ref, filters):
    if filters is None:
        return True

    if ref[0] == '/': ref=ref[1:]
    tmp = ref.split('/')

    for f in filters:
        match = True
        for i,filter_elem in enumerate(f):
            if i >=len(tmp): return True
            ref_elem = tmp[i]
            if (filter_elem != '*') and (ref_elem != filter_elem):
                match = False
                break
        if match: return True
    return False


def prepareFilter(fstr):
    if fstr[0] == '/': fstr=fstr[1:]
    if fstr[-1] == '/': fstr+='*'
    return fstr.split('/')

def parseVersionString(versionString):
    """
    takes a string of the format <major>.<minor>.<patch> (e.g. "3.2.2") and returns a tuple with three integers (major, minor, patch)
    """
    result = pVersion.match(versionString)
    if result is None:
        raise ValueError("VersionString argument did not match the pattern '<major>.<minor>.<patch>'")
    else:
        return (int(result.group(1)),int(result.group(2)),int(result.group(3)))

def findUniqueNameInList(elementList, baseName):
    """
    Attempts to find a unique name in the list of objects based on baseName.
    This function can modify names in gived list.
    Returns a new name which is guaranteed to be unique
    """

    foundElem = None
    highestIndex = 0
    hasIndex = False
    p0 = re.compile(baseName+r'_(\d+)')
    for elem in elementList:
        result = p0.match(elem.name)
        if result is not None:
            hasIndex = True
            index = int(result.group(1))
            if index > highestIndex:
                highestIndex = index
        elif elem.name == baseName:
            foundElem = elem
    if foundElem is not None:
        foundElem.name = '_'.join([foundElem.name, '0'])
    if hasIndex or foundElem is not None:
        return '_'.join([baseName, str(highestIndex+1)])
    else:
        return baseName



class SwDataDefPropsConditional:
    def tag(self,version=None): return 'SW-DATA-DEF-PROPS-CONDITIONAL'
    def __init__(self, baseTypeRef = None, implementationTypeRef = None, swAddressMethodRef = None, swCalibrationAccess = None, swImplPolicy = None, swPointerTargetProps = None, compuMethodRef = None, dataConstraintRef = None, unitRef = None, parent = None):
        self.baseTypeRef = baseTypeRef
        self.swCalibrationAccess = swCalibrationAccess
        self.swAddressMethodRef = swAddressMethodRef
        self.compuMethodRef = compuMethodRef
        self.dataConstraintRef = dataConstraintRef
        self.implementationTypeRef = implementationTypeRef
        self.swPointerTargetProps = swPointerTargetProps
        self.unitRef = unitRef
        self.swImplPolicy = swImplPolicy
        self.parent = parent

    @property
    def swImplPolicy(self):
        return self._swImplPolicy

    @swImplPolicy.setter
    def swImplPolicy(self, value):
        if value is None:
            self._swImplPolicy=None
        else:
            ucvalue=str(value).upper()
            enum_values = ["CONST", "FIXED", "MEASUREMENT-POINT", "QUEUED", "STANDARD"]
            if ucvalue in enum_values:
                self._swImplPolicy = ucvalue
            else:
                raise ValueError('invalid swImplPolicy value: ' +  value)

    def hasAnyProp(self):
        """
        Returns True if any internal attribute is not None, else False.
        The check excludes the parent attribute.
        """
        retval = False
        attr_names = ['baseTypeRef',
                      'swCalibrationAccess',
                      'swAddressMethodRef',
                      'compuMethodRef',
                      'dataConstraintRef',
                      'implementationTypeRef',
                      'swPointerTargetProps',
                      'unitRef',
                      'swImplPolicy'
                      ]
        for name in attr_names:
            if getattr(self, name) is not None:
                retval = True
                break
        return retval

class SwPointerTargetProps:
    """
    (AUTOSAR 4)
    Implements <SW-POINTER-TARGET-PROPS>
    """
    def tag(self, version=None): return 'SW-POINTER-TARGET-PROPS'
    def __init__(self, targetCategory=None, variants = None):
        self.targetCategory = targetCategory
        if variants is None:
            self.variants = []
        else:
            if isinstance(variants, SwDataDefPropsConditional):
                self.variants = [variants]
            else:
                self.variants = list(variants)

class SymbolProps:
    """
    (AUTOSAR 4)
    Implements <SYMBOL-PROPS>
    """
    def tag(self, version=None): return 'SYMBOL-PROPS'

    def __init__(self, name = None, symbol = None):
        self.name = name
        self.symbol = symbol


#Exceptions
class InvalidUnitRef(ValueError):
    pass

class InvalidPortInterfaceRef(ValueError):
    pass

class InvalidComponentTypeRef(ValueError):
    pass

class InvalidDataTypeRef(ValueError):
    pass

class InvalidDataElementRef(ValueError):
    pass

class InvalidPortRef(ValueError):
    pass

class InvalidInitValueRef(ValueError):
    pass

class InvalidDataConstraintRef(ValueError):
    pass

class InvalidCompuMethodRef(ValueError):
    pass

class DataConstraintError(ValueError):
    pass

class InvalidMappingRef(ValueError):
    pass

class InvalidModeGroupRef(ValueError):
    pass

class InvalidModeDeclarationGroupRef(ValueError):
    pass

class InvalidModeDeclarationRef(ValueError):
    pass

class InvalidEventSourceRef(ValueError):
    pass

class InvalidRunnableRef(ValueError):
    pass

class InvalidBehaviorRef(ValueError):
    pass

class InvalidSwAddrmethodRef(ValueError):
    pass



import autosar
import os
import cfile
import sys
from autosar.rte.base import *
from autosar.rte.generator import *
from autosar.rte.partition import *
from autosar.rte.config import *

# class Variable:
#    def __init__(self, name, typename, isQueued=False, queueLen=None, initValue=None):
#       self.name = name
#       self.typename = typename
#       self.hasRead = False
#       self.hasWrite = False
#       self.isQueued = isQueued
#       self.queueLen  = queueLen
#       self.initValue = initValue
# 
# class IntegerVariable(Variable):
#    def __init__(self, name, typename, isQueued=False, queueLen=None, initValue=0):
#       super().__init__(name, typename, isQueued, queueLen, initValue)
#       
# 
# class RecordVariable(Variable):
#    def __init__(self, name, typename, isQueued=False, queueLen=None, initValue=None):
#       super().__init__(name, typename, isQueued, queueLen, initValue)
#       self.elements=[]
# 
# class ArrayVariable(Variable):
#    def __init__(self, name, typename, arrayLen=1, initValue=None):
#       super().__init__(name, typename, initValue=initValue)
#       self.arrayLen=arrayLen
# 
# class BooleanVariable(Variable):
#    def __init__(self, name, typename, initValue=None):
#       super().__init__(name, typename, initValue=initValue)